#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "matplotlib2DViewer.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution
 #  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##
 
__docformat__ = 'restructuredtext'

from fipy.tools import numerix

from fipy.viewers.matplotlibViewer.matplotlibViewer import AbstractMatplotlibViewer, _ColorBar

__all__ = ["Matplotlib2DViewer"]

class AbstractMatplotlib2DViewer(AbstractMatplotlibViewer):
    def figaspect(self, figaspect):
        if figaspect == 'auto':
            figaspect = self.vars[0].mesh.aspect2D
        return figaspect
    
class Matplotlib2DViewer(AbstractMatplotlib2DViewer):
    """
    Displays a contour plot of a 2D `CellVariable` object.    

    The `Matplotlib2DViewer` plots a 2D `CellVariable` using Matplotlib_.

    .. _Matplotlib: http://matplotlib.sourceforge.net/
    """ 
    
    __doc__ += AbstractMatplotlib2DViewer._test2Dirregular(viewer="Matplotlib2DViewer")

    def __init__(self, vars, title=None, limits={}, cmap=None, colorbar='vertical', axes=None, figaspect='auto', **kwlimits):
        """Creates a `Matplotlib2DViewer`.
        

        :Parameters:
          vars
            a `CellVariable` object.
          title
            displayed at the top of the `Viewer` window
          limits : dict
            a (deprecated) alternative to limit keyword arguments
          cmap
            the colormap. Defaults to `matplotlib.cm.jet`
          xmin, xmax, ymin, ymax, datamin, datamax
            displayed range of data. Any limit set to 
            a (default) value of `None` will autoscale.
          colorbar
            plot a colorbar in specified orientation if not `None`
          axes
            if not `None`, `vars` will be plotted into this Matplotlib `Axes` object
          figaspect
            desired aspect ratio of figure. If arg is a number, use that aspect
            ratio. If arg is 'auto', the aspect ratio will be determined from
            the Variable's mesh.
        """
        kwlimits.update(limits)
        AbstractMatplotlib2DViewer.__init__(self, vars=vars, title=title, figaspect=figaspect, 
                                            cmap=cmap, colorbar=colorbar, axes=axes, 
                                            **kwlimits)

        self.mesh = self.vars[0].mesh
        
        vertexIDs = self.mesh._orderedCellVertexIDs

        vertexCoords = self.mesh.vertexCoords

        xCoords = numerix.take(vertexCoords[0], vertexIDs)
        yCoords = numerix.take(vertexCoords[1], vertexIDs)
        
        polys = []

        for x, y in zip(xCoords.swapaxes(0,1), yCoords.swapaxes(0,1)):
            if hasattr(x, 'mask'):
                x = x.compressed()
            if hasattr(y, 'mask'):
                y = y.compressed()
            polys.append(list(zip(x,y)))

        from matplotlib.collections import PolyCollection
        self.collection = PolyCollection(polys)
        self.collection.set_linewidth(0.5)
        try:
            self.axes.add_patch(self.collection)
        except:
            # PolyCollection not child of PatchCollection in matplotlib 0.98
            self.axes.add_collection(self.collection)

        xmin = self._getLimit('xmin', default=xCoords.min())
        xmax = self._getLimit('xmax', default=xCoords.max())
        ymin = self._getLimit('ymin', default=yCoords.min())
        ymax = self._getLimit('ymax', default=yCoords.max())

        self.axes.set_xlim(xmin=xmin, xmax=xmax)
        self.axes.set_ylim(ymin=ymin, ymax=ymax)

        self._plot()
        
    def _getSuitableVars(self, vars):
        from fipy.meshes.mesh2D import Mesh2D
        from fipy.variables.cellVariable import CellVariable
        vars = [var for var in AbstractMatplotlib2DViewer._getSuitableVars(self, vars) \
          if ((var.mesh.dim == 2 and isinstance(var, CellVariable))
              and var.rank == 0)]
        if len(vars) == 0:
            from fipy.viewers import MeshDimensionError
            raise MeshDimensionError("Matplotlib2DViewer can only display a rank-0, 2D CellVariable")
        # this viewer can only display one variable
        return [vars[0]]
        
    def _plot(self):
##         pylab.clf()
        
##         ## Added garbage collection since matplotlib objects seem to hang
##         ## around and accumulate.
##         import gc
##         gc.collect()

        Z = self.vars[0].value 
        
        self.norm.vmin = self._getLimit(('datamin', 'zmin'))
        self.norm.vmax = self._getLimit(('datamax', 'zmax'))

        rgba = self.cmap(self.norm(Z))
        
        self.collection.set_facecolors(rgba)
        self.collection.set_edgecolors(rgba)

        if self.colorbar is not None:
            self.colorbar.plot() #vmin=zmin, vmax=zmax)
        
##        pylab.xlim(xmin=self._getLimit('xmin'),
##                   xmax=self._getLimit('xmax'))

##        pylab.ylim(ymin=self._getLimit('ymin'),
##                   ymax=self._getLimit('ymax'))

if __name__ == "__main__": 
    import fipy.tests.doctestPlus
    fipy.tests.doctestPlus.execButNoTest()

        



#!/usr/bin/env python

## 
 # -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "surfactantConvectionVariable.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

__all__ = ['SurfactantConvectionVariable']

from fipy.tools.numerix import MA
from fipy.tools import numerix

from fipy.tools import vector

from fipy.variables.faceVariable import FaceVariable

class SurfactantConvectionVariable(FaceVariable):
    """
    
    Convection coefficient for the `ConservativeSurfactantEquation`.
    The coeff only has a value for a negative `distanceVar`.

    """

    def __init__(self, distanceVar):
        """
        
        Simple one dimensional test:

        
           >>> from fipy.variables.cellVariable import CellVariable
           >>> from fipy.meshes import Grid2D
           >>> mesh = Grid2D(nx = 3, ny = 1, dx = 1., dy = 1.)
           >>> from fipy.variables.distanceVariable import DistanceVariable
           >>> distanceVar = DistanceVariable(mesh, value = (-.5, .5, 1.5))
           >>> ## answer = numerix.zeros((2, mesh.numberOfFaces),'d')
           >>> answer = FaceVariable(mesh=mesh, rank=1, value=0.).globalValue
           >>> answer[0,7] = -1
           >>> print numerix.allclose(SurfactantConvectionVariable(distanceVar).globalValue, answer)
           True

        Change the dimensions:

           >>> mesh = Grid2D(nx = 3, ny = 1, dx = .5, dy = .25)
           >>> distanceVar = DistanceVariable(mesh, value = (-.25, .25, .75))
           >>> answer[0,7] = -.5
           >>> print numerix.allclose(SurfactantConvectionVariable(distanceVar).globalValue, answer)
           True

        Two dimensional example:

           >>> mesh = Grid2D(nx = 2, ny = 2, dx = 1., dy = 1.)
           >>> distanceVar = DistanceVariable(mesh, value = (-1.5, -.5, -.5, .5))
            >>> answer = FaceVariable(mesh=mesh, rank=1, value=0.).globalValue
           >>> answer[1,2] = -.5
           >>> answer[1,3] = -1
           >>> answer[0,7] = -.5
           >>> answer[0,10] = -1
           >>> print numerix.allclose(SurfactantConvectionVariable(distanceVar).globalValue, answer)
           True

        Larger grid:

           >>> mesh = Grid2D(nx = 3, ny = 3, dx = 1., dy = 1.)
           >>> distanceVar = DistanceVariable(mesh, value = (1.5, .5 , 1.5,
           ...                                           .5 , -.5, .5 ,
           ...                                           1.5, .5 , 1.5))
            >>> answer = FaceVariable(mesh=mesh, rank=1, value=0.).globalValue
           >>> answer[1,4] = .25
           >>> answer[1,7] = -.25
           >>> answer[0,17] = .25
           >>> answer[0,18] = -.25
           >>> print numerix.allclose(SurfactantConvectionVariable(distanceVar).globalValue, answer)
           True
           
        """
        
        FaceVariable.__init__(self, mesh=distanceVar.mesh, name='surfactant convection', rank=1)
        self.distanceVar = self._requires(distanceVar)

    def _calcValue(self):

        Nfaces = self.mesh.numberOfFaces
        M = self.mesh._maxFacesPerCell
        dim = self.mesh.dim
        cellFaceIDs = self.mesh.cellFaceIDs
     
        faceNormalAreas = self.distanceVar._levelSetNormals * self.mesh._faceAreas

        cellFaceNormalAreas = numerix.array(MA.filled(numerix.take(faceNormalAreas, cellFaceIDs, axis=-1), 0))
        norms = numerix.array(MA.filled(MA.array(self.mesh._cellNormals), 0))
        
        alpha = numerix.dot(cellFaceNormalAreas, norms)
        alpha = numerix.where(alpha > 0, alpha, 0)

        alphasum = numerix.sum(alpha, axis=0)
        alphasum += (alphasum < 1e-100) * 1.0
        alpha = alpha / alphasum

        phi = numerix.repeat(self.distanceVar[numerix.newaxis, ...], M, axis=0)
        alpha = numerix.where(phi > 0., 0, alpha)
        
        volumes = numerix.array(self.mesh.cellVolumes)
        alpha = alpha * volumes * norms

        value = numerix.zeros((dim, Nfaces),'d')

        vector._putAdd(value, cellFaceIDs, alpha, mask=MA.getmask(MA.array(cellFaceIDs)))

##         value = numerix.reshape(value, (dim, Nfaces, dim))

        return -value / self.mesh._faceAreas

def _test(): 
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()
    
if __name__ == "__main__": 
    _test() 



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "linearLUSolver.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

import os

from pysparse import superlu

from fipy.solvers.pysparse.pysparseSolver import PysparseSolver
from fipy.tools import numerix

DEBUG = False

__all__ = ["LinearLUSolver"]

class LinearLUSolver(PysparseSolver):
    """
    
    The `LinearLUSolver` solves a linear system of equations using
    LU-factorisation. This method solves systems with a general
    non-symmetric coefficient matrix using partial pivoting.

    The `LinearLUSolver` is a wrapper class for the the PySparse_
    `superlu.factorize()` method.

    .. _PySparse: http://pysparse.sourceforge.net
    
    """
    
    def __init__(self, tolerance=1e-10, iterations=10,
                       maxIterations=10, precon=None):
        """
        Creates a `LinearLUSolver`.

        :Parameters:
          - `tolerance`: The required error tolerance.
          - `iterations`: The number of LU decompositions to perform.
            For large systems a number of iterations is generally required.
          - `precon`: not used but maintains a common interface.
          
        """

        iterations = min(iterations, maxIterations)
        
        super(LinearLUSolver, self).__init__(tolerance = tolerance, 
                                             iterations = iterations)

    def _solve_(self, L, x, b):
        diag = L.takeDiagonal()
        maxdiag = max(numerix.absolute(diag))

        L = L * (1 / maxdiag)
        b = b * (1 / maxdiag)

        LU = superlu.factorize(L.matrix.to_csr())

        if DEBUG:
            import sys
            print(L.matrix, file=sys.stderr)

        error0 = numerix.sqrt(numerix.sum((L * x - b)**2))

        for iteration in range(self.iterations):
            errorVector = L * x - b

            if (numerix.sqrt(numerix.sum(errorVector**2)) / error0)  <= self.tolerance:
                break

            xError = numerix.zeros(len(b),'d')
            LU.solve(errorVector, xError)
            x[:] = x - xError
            
        if 'FIPY_VERBOSE_SOLVER' in os.environ:
            from fipy.tools.debug import PRINT        
            PRINT('iterations: %d / %d' % (iteration+1, self.iterations))
            PRINT('residual:', numerix.sqrt(numerix.sum(errorVector**2)))




#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "modPhysicalField.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  
 #  Description: 
 # 
 # Physical fields or quantities with units
 #
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

__all__ = []

from fipy.tools.dimensions.physicalField import PhysicalField

from fipy.tools import numerix

class _ModPhysicalField(PhysicalField):

    def mod(argument):
        return numerix.fmod(argument + 3. * numerix.pi, 2. * numerix.pi) - numerix.pi
    mod = staticmethod(mod)

    def __sub__(self, other):
        if isinstance(other, _ModPhysicalField):
            return self.__class__(value= self.mod(self.inRadians() - other.inRadians()), unit="rad")
        else:
            return self._sum(other, sign2 = lambda b: -b)
    
    def __rsub__(self, other):
        if isinstance(other, _ModPhysicalField):
            return self.__class__(value = self.mod(argument=other.inRadians() - self.inRadians()), unit="rad")
        else:
            return self._sum(other, sign1 = lambda a: -a)

    def ravel(self):
        return self.value.ravel()



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "dump.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution
 #  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

import pickle
import os
import sys
import gzip

from fipy.tools import parallelComm

__all__ = ["write", "read"]

# TODO: add test to show that round trip pickle of mesh doesn't work properly
# FIXME: pickle fails to work properly on numpy 1.1 (run gapFillMesh.py)
def write(data, filename = None, extension = '', communicator=parallelComm):
    """
    Pickle an object and write it to a file. Wrapper for
    `cPickle.dump()`.

    :Parameters:
      - `data`: The object to be pickled.
      - `filename`: The name of the file to place the pickled object. If `filename` is `None`
        then a temporary file will be used and the file object and file name will be returned as a tuple
      - `extension`: Used if filename is not given.
      - `communicator`: Object with `procID` and `Nproc` attributes.

    Test to check pickling and unpickling.

        >>> from fipy.meshes import Grid1D
        >>> old = Grid1D(nx = 2)
        >>> f, tempfile = write(old)
        >>> new = read(tempfile, f)
        >>> print old.numberOfCells == new.numberOfCells
        True
        
    """
    if communicator.procID == 0:
        if filename is None:
            import tempfile
            (f, _filename) =  tempfile.mkstemp(extension)
        else:
            (f, _filename) = (None, filename)
        fileStream = gzip.GzipFile(filename = _filename, mode = 'w', fileobj = None)
    else:
        fileStream = open(os.devnull, mode='w')
        (f, _filename) = (None, os.devnull)
        
    pickle.dump(data, fileStream, 0)
    fileStream.close()
        
    if filename is None:
        return (f, _filename)

def read(filename, fileobject=None, communicator=parallelComm, mesh_unmangle=False):
    """
    Read a pickled object from a file. Returns the unpickled object.
    Wrapper for `cPickle.load()`.

    :Parameters:
      - `filename`: The name of the file to unpickle the object from.
      - `fileobject`: Used to remove temporary files
      - `communicator`: Object with `procID` and `Nproc` attributes.
      - `mesh_unmangle`: Correct improper pickling of non-uniform meshes (ticket:243)
      
    """
    if communicator.procID == 0:
        fileStream = gzip.GzipFile(filename = filename, mode = 'r', fileobj = None)
        data = fileStream.read()
        fileStream.close()
        if fileobject is not None:
            os.close(fileobject)
            os.remove(filename)
    else:
        data = None
        
    if communicator.Nproc > 1:
        data = communicator.bcast(data, root=0)

    if sys.version_info < (3,0):
        import io
        f = io.StringIO(data)
    else:
        import io
        f = io.BytesIO(data)
        
    unpickler = pickle.Unpickler(f)
    
    if mesh_unmangle:
        def find_class(module, name):
            __import__(module)
            mod = sys.modules[module]
            klass = getattr(mod, name)
            
            from fipy import meshes
            import types
            
            if isinstance(klass, type) and issubclass(klass, meshes.mesh.Mesh):
                class UnmangledMesh(klass):
                    def __setstate__(self, dict):
                        if ('cellFaceIDs' in dict 
                            and 'faceVertexIDs' in dict):
                                
                            dict = dict.copy()
                            for key in ('cellFaceIDs', 'faceVertexIDs'):
                                arr = dict[key]
                                arr.data[:] = arr.transpose().flatten().reshape(arr.shape)
                            
                        klass.__setstate__(self, dict)
                    
                return UnmangledMesh
            else:
                return klass

        unpickler.find_global = find_class
        
    return unpickler.load()

def _test(): 
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()
    
if __name__ == "__main__": 
    _test()     



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "variable.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

import os

from fipy.tools.dimensions import physicalField
from fipy.tools import numerix
from fipy.tools import parser
from fipy.tools import inline
from fipy.tools.decorators import getsetDeprecated, mathMethodDeprecated

__all__ = ["Variable"]

class Variable(object):
    """
    Lazily evaluated quantity with units. 

    Using a :class:`~fipy.variables.variable.Variable` in a mathematical expression will create an
    automatic dependency :class:`~fipy.variables.variable.Variable`, e.g.,

    >>> a = Variable(value=3)
    >>> b = 4 * a
    >>> b
    (Variable(value=array(3)) * 4)
    >>> b()
    12
        
    Changes to the value of a :class:`~fipy.variables.variable.Variable` will automatically trigger
    changes in any dependent :class:`~fipy.variables.variable.Variable` objects

    >>> a.setValue(5)
    >>> b
    (Variable(value=array(5)) * 4)
    >>> print b()
    20
    """

    _cacheAlways = (os.getenv("FIPY_CACHE") is not None) or False
    if parser.parse("--no-cache", action="store_true"):
        _cacheAlways = False
    if parser.parse("--cache", action="store_true"):
        _cacheAlways = True

    _cacheNever = False
    
    def __new__(cls, *args, **kwds):
        return object.__new__(cls)
    
    def __init__(self, value=0., unit=None, array=None, name='', cached=1):
        """
        Create a `Variable`.
        
            >>> Variable(value=3)
            Variable(value=array(3))
            >>> Variable(value=3, unit="m")
            Variable(value=PhysicalField(3,'m'))
            >>> Variable(value=3, unit="m", array=numerix.zeros((3,2), 'l'))
            Variable(value=PhysicalField(array([[3, 3],
                   [3, 3],
                   [3, 3]]),'m'))

        :Parameters:
          - `value`: the initial value
          - `unit`: the physical units of the `Variable`
          - `array`: the storage array for the `Variable`
          - `name`: the user-readable name of the `Variable`
          - `cached`: whether to cache or always recalculate the value
        """
            
        self.requiredVariables = []
        self.subscribedVariables = []

        if isinstance(value, Variable):
            value = value.value
            if hasattr(value, 'copy'):
                value = value.copy()
            unit = None
            
        self._setValueInternal(value=value, unit=unit, array=array)
        
        self._name = name
                
        self._cached = cached

        self.stale = 1
        self._markFresh()
        
##    __array_priority__ and __array_wrap__ are required to override
##    the default behavior of numpy. If a numpy array and a Variable
##    are in a binary operation and numpy is first, then numpy will,
##    by default, try and do everything it can to get a a raw numpy
##    array out of Variable. __array_wrap__ seems to have been
##    introduced into masked array to fix this issue. __array_wrap__ is
##    called after the operation is done so it could hurt efficiency badly.
##    Something else needs to be done to stop the initial evaluation.

    __array_priority__ = 100.0    

    def __array_wrap__(self, arr, context=None):
        """
        Required to prevent numpy not calling the reverse binary operations.
        Both the following tests are examples ufuncs.
        
        >>> print type(numerix.array([1.0, 2.0]) * Variable([1.0, 2.0]))
        <class 'fipy.variables.binaryOperatorVariable.binOp'>

        >>> from scipy.special import gamma as Gamma # doctest: +SCIPY
        >>> print type(Gamma(Variable([1.0, 2.0]))) # doctest: +SCIPY
        <class 'fipy.variables.unaryOperatorVariable.unOp'>
        """
        result = arr
        
        if context is not None:
            from fipy.variables.constant import _Constant
            (func, args, _) = context
            def __makeVariable(v):
                if not isinstance(v, Variable):
                    v = _Constant(v)
                return v
            args = [__makeVariable(arg) for arg in args]

            cannotInline = ["expi", "logical_and", "logical_or", "logical_not", "logical_xor", "sign", 
                            "conjugate", "dot", "allclose", "allequal"]
            if len(args) == 1:
                result = args[0]._UnaryOperatorVariable(op=func, opShape=arr.shape, canInline=func.__name__ not in cannotInline)
            elif len(args) == 2:
                result = args[0]._BinaryOperatorVariable(op=func, other=args[1], opShape=arr.shape, canInline=func.__name__ not in cannotInline)
            else:
                result = NotImplemented

        return result
        
    def __array__(self, t=None):
        """
        Attempt to convert the `Variable` to a numerix `array` object
    
            >>> v = Variable(value=[2,3])
            >>> print numerix.array(v)
            [2 3]
        
        A dimensional `Variable` will convert to the numeric value in its base units
    
            >>> v = Variable(value=[2,3], unit="mm")
            >>> numerix.array(v)
            array([ 0.002,  0.003])
        """

        return numerix.array(self.value, t)

##    def _get_array_interface(self):
##        return self._array.__array_interface__
     
##    def _set_array_interface(self, value):
##        self._array.__array_interface__ = value
         
##    def _del_array_interface(self):
##        del self._array.__array_interface__
  
##    __array_interface__ = property(_get_array_interface,
##                                   _set_array_interface,
##                                   _del_array_interface,
##                                   "the '__array_inteface__'")
        
    def copy(self):
        """
        Make an duplicate of the `Variable`
        
            >>> a = Variable(value=3)
            >>> b = a.copy()
            >>> b
            Variable(value=array(3))

        The duplicate will not reflect changes made to the original
                          
            >>> a.setValue(5)
            >>> b
            Variable(value=array(3))
            
        Check that this works for arrays.
        
            >>> a = Variable(value=numerix.array((0,1,2)))
            >>> b = a.copy()
            >>> b
            Variable(value=array([0, 1, 2]))
            >>> a[1] = 3
            >>> b
            Variable(value=array([0, 1, 2]))
            
        """
        return self._getArithmeticBaseClass()(value=self)


    @getsetDeprecated
    def _getUnitAsOne(self):
        return self._unitAsOne

    @property
    def _unitAsOne(self):
        unit = self.unit
        if unit is physicalField._unity:
            return 1.
        else:
            return physicalField.PhysicalField(value=1, unit=unit)

    def _extractUnit(self, value):
        if isinstance(value, physicalField.PhysicalField):
            return value.unit
        else:
            return physicalField._unity 

    @getsetDeprecated
    def getUnit(self):
        return self.unit

    def _getUnit(self):
        """
        Return the unit object of `self`.
        
            >>> Variable(value="1 m").unit
            <PhysicalUnit m>
        """
        return self._extractUnit(self.value)
        
    @getsetDeprecated
    def setUnit(self, unit):
        self.unit = unit

    def _setUnit(self, unit):
        """
        Change the unit object of `self` to `unit`
        
            >>> a = Variable(value="1 m")
            >>> a.unit = "m**2/s"
            >>> print a
            1.0 m**2/s
        """
        if self._value is None:
            self.value

        if isinstance(self._value, physicalField.PhysicalField):
            self._value.unit = unit
        else:
            self._value = physicalField.PhysicalField(value=self._value, unit=unit)

    unit = property(_getUnit, _setUnit)

    def inBaseUnits(self):
        """
        Return the value of the `Variable` with all units reduced to 
        their base SI elements.
        
            >>> e = Variable(value="2.7 Hartree*Nav")
            >>> print e.inBaseUnits().allclose("7088849.01085 kg*m**2/s**2/mol")
            1
        """
        value = self.value
        if isinstance(value, physicalField.PhysicalField):
            return value.inBaseUnits()
        else:
            return value

    def inUnitsOf(self, *units):
        """
        Returns one or more `Variable` objects that express the same
        physical quantity in different units.  The units are specified by
        strings containing their names.  The units must be compatible with
        the unit of the object.  If one unit is specified, the return value
        is a single `Variable`.
        
        >>> freeze = Variable('0 degC')
        >>> print freeze.inUnitsOf('degF').allclose("32.0 degF")
        1
        
        If several units are specified, the return value is a tuple of
        `Variable` instances with with one element per unit such that
        the sum of all quantities in the tuple equals the the original
        quantity and all the values except for the last one are integers.
        This is used to convert to irregular unit systems like
        hour/minute/second.  The original object will not be changed.
        
        >>> t = Variable(value=314159., unit='s')
        >>> print numerix.allclose([e.allclose(v) for (e, v) in zip(t.inUnitsOf('d','h','min','s'),
        ...                                                         ['3.0 d', '15.0 h', '15.0 min', '59.0 s'])], 
        ...                        True)
        1
        """
        value = self.value
        if isinstance(value, physicalField.PhysicalField):
            return value.inUnitsOf(*units)
        else:
            return value

##     def __getitem__(self, index):
##         """    
##         "Evaluate" the `Variable` and return the specified element
##       
##            >>> ## a = Variable(value=((3.,4.),(5.,6.)), unit="m") + "4 m"
##            >>> ## print a[1,1]
##             10.0 m
## 
##         It is an error to slice a `Variable` whose `value` is not sliceable
## 
##            >>> ## Variable(value=3)[2]
##             Traceback (most recent call last):
##                   ...
##             IndexError: 0-d arrays can't be indexed
## 
##         """
##         return (self.value)[index]
                            
    @getsetDeprecated
    def getName(self):
        return self.name
        
    @getsetDeprecated
    def setName(self, name):
        self.name = name

    def _getName(self):
        return self._name

    def _setName(self, name):
        self._name = name

    name = property(_getName, _setName)
    
    def __str__(self):
        return str(self.value)
            
    def __repr__(self):
        if hasattr(self, 'name') and len(self.name) > 0:
            return self.name
        else:
            s = self.__class__.__name__ + '('
            s += 'value=' + repr(self.value)
            s += ')'
            return s

    def _getCIndexString(self, shape):
        r"""
        Test for inline issue. (1, ni) shapes were not handled correctly.

        >>> from fipy import *
        >>> mesh = Tri2D(dx=1., dy=1., nx=1, ny=1)
        >>> diffCoeff = FaceVariable(mesh = mesh, value = 1.0)
        >>> normals = FaceVariable(mesh=mesh, rank=1, value=mesh._orientedFaceNormals)
        >>> normalsNthCoeff = diffCoeff[numerix.newaxis] * normals 

        First variable value access does not provoke inlining.

        >>> value = normalsNthCoeff.value
        >>> print (normalsNthCoeff == value).all()
        True

        """

        dimensions = len(shape)
        if dimensions == 1:
            return '[i]'
        elif dimensions == 2:
            if shape[-1] == 1:
                return '[j]'
            elif shape[0] == 1:
                return '[i]'
            else:
                return '[i + j * ni]'
        elif dimensions == 3:
            if shape[-1] == 1:
                if shape[-2] == 1:
                    return '[k]'
                else:
                    return '[j + k * nj]'
            elif shape[-2] == 1:
                return '[i + k * ni * nj]'
            else:
                return '[i + j * ni + k * ni * nj]'
            

    def _getCstring(self, argDict={}, id="", freshen=None):
         """
         Generate the string and dictionary to be used in inline
             >>> (Variable((1)))._getCstring(argDict={})
             'var'
           
             >>> (Variable((1,2,3,4)))._getCstring(argDict={})
             'var[i]'
       
             >>> (Variable(((1,2),(3,4))))._getCstring(argDict={})
             'var[i + j * ni]'

             >>> Variable((((1,2),(3,4)),((5,6),(7,8))))._getCstring(argDict={})
             'var[i + j * ni + k * ni * nj]'

             >>> (Variable(1) * Variable((1,2,3)))._getCstring(argDict={})
             '(var0 * var1[i])'

         freshen is ignored

         Testing when a cell variable multiplies an array that has a
         shape, but has olny one element. This works regullarly,
         but fails when inlining.

         >>> from fipy import *
         >>> m = Grid1D(nx=3)
         >>> x = m.cellCenters[0]
         >>> tmp = m.cellCenters[0] * numerix.array(((0.,), (1.,)))[1]
         >>> print numerix.allclose(tmp, x)
         True
         >>> print numerix.allclose(tmp, x)
         True
         
         """
         
         identifier = 'var%s' % (id)

         v = self.value

         if type(v) not in (type(numerix.array(1)),):
             varray = numerix.array(v)
         else:
             varray = v

         if len(varray.shape) == 0:
             if varray.dtype in (numerix.array(1).dtype,):
                 argDict[identifier] = int(varray)
             elif varray.dtype in (numerix.array(1.).dtype,):
                 argDict[identifier] = float(varray)
             else:
                 argDict[identifier] = varray
         else:
             argDict[identifier] = varray
             
         try:
             shape = self.opShape
         except AttributeError:
             shape = self.shape

         if len(shape) == 0:
             return identifier
         elif len(shape) == 1 and shape[0] == 1:
             return identifier + '[0]'
         else:
             return identifier + self._getCIndexString(shape)

    def tostring(self, max_line_width=75, precision=8, suppress_small=False, separator=' '):
        return numerix.tostring(self.value, 
                                max_line_width=max_line_width,
                                precision=precision, 
                                suppress_small=suppress_small, 
                                separator=separator)
        
    def __setitem__(self, index, value):
        if self._value is None:
            self._getValue()
        self._value[index] = value
        self._markFresh()
        
    def itemset(self, value):
        if self._value is None:
            self._getValue()
        self._value.itemset(value)
        self._markFresh()
        
    def put(self, indices, value):
        if self._value is None:
            self._getValue()
        numerix.put(self._value, indices, value)
        self._markFresh()
        
    def __call__(self):
        """
        "Evaluate" the `Variable` and return its value
        
            >>> a = Variable(value=3)
            >>> print a()
            3
            >>> b = a + 4
            >>> b
            (Variable(value=array(3)) + 4)
            >>> b()
            7
        """
        return self.value

    @getsetDeprecated
    def getValue(self):
        return self.value

    def _getValue(self):
        """
        "Evaluate" the `Variable` and return its value (longhand)
        
            >>> a = Variable(value=3)
            >>> print a.value
            3
            >>> b = a + 4
            >>> b
            (Variable(value=array(3)) + 4)
            >>> b.value
            7

        """
        
        if self.stale or not self._isCached() or self._value is None:
            value = self._calcValue()
            if self._isCached():
                self._setValueInternal(value=value)
            else:
                self._setValueInternal(value=None)
            self._markFresh()
        else:
            value = self._value

        if len(self.constraints) > 0:
            value = value.copy()
            for constraint in self.constraints:
                if constraint.where is None:
                    value[:] = constraint.value
                else:
                    mask = constraint.where
                    if not hasattr(mask, 'dtype') or mask.dtype != bool:
                        mask = numerix.array(mask, dtype=numerix.NUMERIX.bool)

                    if 0 not in value.shape:
                        try:
                            value[..., mask] = constraint.value
                        except:
                            value[..., mask] = numerix.array(constraint.value)[..., mask]

        return value

    def _setValueProperty(self, newVal):
        """Since `self.setValue` contains optional, named parameters, we will
        punt the property's set method off to that."""
        self.setValue(newVal)

    value = property(_getValue, _setValueProperty)
    
    @property
    def constraints(self):
        if not hasattr(self, "_constraints"):
            self._constraints = []
        return self._constraints
            
    def constrain(self, value, where=None):
        """
        Constrain the `Variable` to have a `value` at an index or mask location specified by `where`.

        >>> v = Variable((0,1,2,3))
        >>> v.constrain(2, numerix.array((True, False, False, False)))
        >>> print v
        [2 1 2 3]
        >>> v[:] = 10
        >>> print v
        [ 2 10 10 10]
        >>> v.constrain(5, numerix.array((False, False, True, False)))
        >>> print v
        [ 2 10  5 10]
        >>> v[:] = 6
        >>> print v
        [2 6 5 6]
        >>> v.constrain(8)
        >>> print v
        [8 8 8 8]
        >>> v[:] = 10
        >>> print v
        [8 8 8 8]
        >>> del v.constraints[2]
        >>> print v
        [ 2 10  5 10]

        >>> from fipy.variables.cellVariable import CellVariable
        >>> from fipy.meshes import Grid2D
        >>> m = Grid2D(nx=2, ny=2)
        >>> x, y = m.cellCenters
        >>> v = CellVariable(mesh=m, rank=1, value=(x, y))
        >>> v.constrain(((0.,), (-1.,)), where=m.facesLeft)
        >>> print v.faceValue
        [[ 0.5  1.5  0.5  1.5  0.5  1.5  0.   1.   1.5  0.   1.   1.5]
         [ 0.5  0.5  1.   1.   1.5  1.5 -1.   0.5  0.5 -1.   1.5  1.5]]
        
        :Parameters:
          - `value`: the value of the constraint
          - `where`: the constraint mask or index specifying the location of the constraint

        """

        from fipy.boundaryConditions.constraint import Constraint
        if not isinstance(value, Constraint):
            value = Constraint(value=value, where=where)
            
        if not hasattr(self, "_constraints"):
            self._constraints = []
        self._constraints.append(value)
        self._requires(value.value)
        self._markStale()
        
    def release(self, constraint):
        """Remove `constraint` from `self`
        
        >>> v = Variable((0,1,2,3))
        >>> v.constrain(2, numerix.array((True, False, False, False)))
        >>> v[:] = 10
        >>> from fipy.boundaryConditions.constraint import Constraint
        >>> c1 = Constraint(5, numerix.array((False, False, True, False)))
        >>> v.constrain(c1)
        >>> v[:] = 6
        >>> v.constrain(8)
        >>> v[:] = 10
        >>> del v.constraints[2]
        >>> v.release(constraint=c1)
        >>> print v
        [ 2 10 10 10]
        """
        self.constraints.remove(constraint)
        
    def _isCached(self):
        return self._cacheAlways or (self._cached and not self._cacheNever)
        
    def cacheMe(self, recursive=False):
        self._cached = True
        if recursive:
            for var in self.requiredVariables:
                var.cacheMe(recursive=True)
                
    def dontCacheMe(self, recursive=False):
        self._cached = False
        if recursive:
            for var in self.requiredVariables:
                var.dontCacheMe(recursive=False)

    def _setValueInternal(self, value, unit=None, array=None):
        self._value = self._makeValue(value=value, unit=unit, array=array)
     
    def _makeValue(self, value, unit=None, array=None):

        ## --inline code often returns spurious results with noncontiguous
        ## arrays. A test case was put in _execInline(). The best fix turned out to
        ## be here.
        
        if (inline.doInline 
            and hasattr(value, 'iscontiguous') and not value.iscontiguous()):
            value = value.copy()
            
        if isinstance(value, Variable):
            value = value.value
            
        PF = physicalField.PhysicalField

        if not isinstance(value, PF):
            
            if getattr(self, '_value', None) is not None:
                v = self._value
                if isinstance(v, PF):
                    v = self._value.value
                if type(value) in (type(1), type(1.)):
                    if type(v) is type(numerix.array(1)):
                        if v.shape is not ():
##                        if len(v) > 1:
                            value = numerix.resize(value, v.shape).astype(v.dtype)
                    
            if unit is not None or type(value) in [type(''), type(()), type([])]:
                value = PF(value=value, unit=unit, array=array)
            elif array is not None:
                array[:] = value
                value = array
            elif type(value) not in (type(None), type(numerix.array(1)), type(numerix.MA.array(1))):
                value = numerix.array(value)
##                 # numerix does strange things with really large integers.
##                 # Even though Python knows how to do arithmetic with them,
##                 # Numeric converts them to 'O' objects that it then doesn't understand.
##                 if value.typecode() == 'O':
##                     value = numerix.array(float(value))

        if isinstance(value, PF) and value.unit.isDimensionless():
            value = value.numericValue
            
        return value

    def setValue(self, value, unit=None, where=None):
        """
        Set the value of the Variable. Can take a masked array.

            >>> a = Variable((1,2,3))
            >>> a.setValue(5, where=(1, 0, 1))
            >>> print a
            [5 2 5]

            >>> b = Variable((4,5,6))
            >>> a.setValue(b, where=(1, 0, 1))
            >>> print a
            [4 2 6]
            >>> print b
            [4 5 6]
            >>> a.value = 3
            >>> print a
            [3 3 3]

            >>> b = numerix.array((3,4,5))
            >>> a.value = b
            >>> a[:] = 1
            >>> print b
            [3 4 5]

            >>> a.setValue((4,5,6), where=(1, 0)) #doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ....
            ValueError: shape mismatch: objects cannot be broadcast to a single shape
            
        """
        if where is not None:
            tmp = numerix.empty(numerix.getShape(where), self.getsctype())
            tmp[:] = value
            tmp = numerix.where(where, tmp, self.value)
        else:
            if hasattr(value, 'copy'):
                tmp = value.copy()
            else:
                tmp = value

        value = self._makeValue(value=tmp, unit=unit, array=None)

        if numerix.getShape(self._value) == ():
            self._value.itemset(value)
        else:
            self._value[:] = value
            
        self._markFresh()
        
    def _setNumericValue(self, value):
        if isinstance(self._value, physicalField.PhysicalField):
            self._value.value = value
        else:
            self._value = value
        
    @getsetDeprecated
    def _getArray(self):
        return self._array

    @property
    def _array(self):
        if isinstance(self._value, physicalField.PhysicalField):
            return self._value._array
        else:
            return self._value
            
    @getsetDeprecated
    def getNumericValue(self):
        return self.numericValue

    @property
    def numericValue(self):
        value = self.value
        if isinstance(value, physicalField.PhysicalField):
            return value.numericValue
        else:
            return value
            
    @getsetDeprecated
    def getShape(self):
        return self.shape

    def _getShape(self):
        """
        Tuple of array dimensions.

        >>> Variable(value=3).shape
        ()
        >>> Variable(value=(3,)).shape
        (1,)
        >>> Variable(value=(3,4)).shape
        (2,)
        
        >>> Variable(value="3 m").shape
        ()
        >>> Variable(value=(3,), unit="m").shape
        (1,)
        >>> Variable(value=(3,4), unit="m").shape
        (2,)
        """
        if self._value is not None:
            return numerix.getShape(self._value)
        else:
            return ()
    
    shape = property(_getShape)

    def getsctype(self, default=None):
        """

        Returns the Numpy sctype of the underlying array.

            >>> Variable(1).getsctype() == numerix.NUMERIX.obj2sctype(numerix.array(1))
            True
            >>> Variable(1.).getsctype() == numerix.NUMERIX.obj2sctype(numerix.array(1.))
            True
            >>> Variable((1,1.)).getsctype() == numerix.NUMERIX.obj2sctype(numerix.array((1., 1.)))
            True
            
        """
        
        if not hasattr(self, 'typecode'):
            self.typecode = numerix.obj2sctype(rep=self.numericValue, default=default)
        
        return self.typecode
        
    @property
    def itemsize(self):
        return self.value.itemsize
        
    def _calcValue(self):
        return self._value

    def _calcValueNoInline(self):
        raise NotImplementedError

    def _calcValueInline(self):
        raise NotImplementedError
    
    @getsetDeprecated
    def getSubscribedVariables(self):
        return self.subscribedVariables

    def _getSubscribedVariables(self):
        self._subscribedVariables = [sub for sub in self._subscribedVariables if sub() is not None]
        
        return self._subscribedVariables

    def _setSubscribedVariables(self, sVars):
        self._subscribedVariables = sVars

    subscribedVariables = property(_getSubscribedVariables,
                                   _setSubscribedVariables)
        
    def __markStale(self):
        for subscriber in self.subscribedVariables:
            if subscriber() is not None:
                ## Even though getSubscribedVariables() strips out dead 
                ## references, subscriber() might still be dead due to the 
                ## vagaries of garbage collection and the possibility that 
                ## later subscribedVariables were removed, changing the 
                ## dependencies of this subscriber. 
                ## See <https://github.com/usnistgov/fipy/issues/103> for more explanation.
                subscriber()._markStale()
                
    def _markFresh(self):
        self.stale = 0
        self.__markStale()

    def _markStale(self):
        if not self.stale:
            self.stale = 1
            self.__markStale()
            
    def _requires(self, var):
        if isinstance(var, Variable):
            self.requiredVariables.append(var)
            var._requiredBy(self)
        else:
            from fipy.variables.constant import _Constant
            var = _Constant(value=var)
        self._markStale()
        return var
            
    def _requiredBy(self, var):
        assert isinstance(var, Variable)
        
        # we retain a weak reference to avoid a memory leak 
        # due to circular references between the subscriber
        # and the subscribee
        import weakref
        self.subscribedVariables.append(weakref.ref(var))
        
    @getsetDeprecated
    def _getVariableClass(self):
        return self._variableClass

    @property
    def _variableClass(self):
        return Variable
        
    def _execInline(self, comment=None):
        """
        Gets the stack from _getCstring() which calls _getRepresentation()
        
            >>> (Variable((1,2,3,4)) * Variable((5,6,7,8)))._getCstring()
            '(var0[i] * var1[i])'
            >>> (Variable(((1,2),(3,4))) * Variable(((5,6),(7,8))))._getCstring()
            '(var0[i + j * ni] * var1[i + j * ni])'
            >>> (Variable((1,2)) * Variable((5,6)) * Variable((7,8)))._getCstring()
            '((var00[i] * var01[i]) * var1[i])'

        The following test was implemented due to a problem with
        contiguous arrays.  The `mesh.cellCenters[1]` command
        introduces a non-contiguous array into the `Variable` and this
        causes the inline routine to return senseless results.
        
            >>> from fipy import Grid2D, CellVariable
            >>> mesh = Grid2D(dx=1., dy=1., nx=2, ny=2)
            >>> var = CellVariable(mesh=mesh, value=0.)
            >>> Y =  mesh.cellCenters[1]
            >>> var.value = (Y + 1.0)
            >>> print var - Y
            [ 1.  1.  1.  1.]
        """
    
        from fipy.tools import inline
        argDict = {}
        string = self._getCstring(argDict=argDict, freshen=True) + ';'
        
        try:
            shape = self.opShape
        except AttributeError:
            shape = self.shape

        dimensions = len(shape)
            
        if dimensions == 0:
            string = 'result[0] = ' + string
            dim = ()
        else:
            string = 'result' + self._getCIndexString(shape) + ' = ' + string
            ni = self.opShape[-1]
            argDict['ni'] = ni
            if dimensions == 1:
                dim = (ni)
            else:
                nj = self.opShape[-2]
                argDict['nj'] = nj
                if dimensions == 2:
                    dim =(nj,ni)
                elif dimensions == 3:
                    nk = self.opShape[-3]
                    dim = (nk,nj,ni)
                    argDict['nk'] = nk
                else:
                    raise DimensionError('Impossible Dimensions')

        ## Following section makes sure that the result array has a
        ## valid typecode. If self._value is None then a typecode is
        ## assigned to the Variable by running the calculation without
        ## inlining. The non-inlined result is thus used the first
        ## time through.

        
        if self._value is None and not hasattr(self, 'typecode'):
            self.canInline = False
            argDict['result'] = self.value
            self.canInline = True
            self.typecode = numerix.obj2sctype(argDict['result'])
        else:
            if self._value is None:
                if self.getsctype() == numerix.bool_:
                    argDict['result'] = numerix.empty(dim, numerix.int8)
                else:
                    argDict['result'] = numerix.empty(dim, self.getsctype())
            else:
                argDict['result'] = self._value

            resultShape = argDict['result'].shape

            if resultShape == ():
                argDict['result'] = numerix.reshape(argDict['result'], (1,))

            inline._runInline(string, converters=None, comment=comment, **argDict)

            if resultShape == ():
                argDict['result'] = numerix.reshape(argDict['result'], resultShape)
                
            if self.getsctype() == numerix.bool_:
                argDict['result'] = numerix.asarray(argDict['result'], dtype=self.getsctype())

        return argDict['result']

    def _broadcastShape(self, other):
        ignore, ignore, broadcastshape = numerix._broadcastShapes(self.shape, numerix.getShape(other))
        
        return broadcastshape
        
##         selfshape = self.shape
##         othershape = other.shape
##         
##         if len(selfshape) > len(othershape):
##             othershape = (1,) * (len(selfshape) - len(othershape)) + othershape
##         elif len(selfshape) < len(othershape):
##             selfshape = (1,) * (len(othershape) - len(selfshape)) + selfshape
##         
##         if numerix.logical_and.reduce([(s == o or s == 1 or o == 1) for s,o in zip(selfshape, othershape)]):
##             return tuple([max(s,o) for s,o in zip(selfshape, othershape)])
##         else:
##             return None
            
    def _getArithmeticBaseClass(self, other=None):
        """
        Given `self` and `other`, return the desired base class for an operation
        result.
        """
        if other is None:
            return Variable
            
        if self._broadcastShape(other) is not None:
            # If self and other have the same base class, result has that base class.
            # If self derives from other, result has self's base class.
            # If other derives from self, result has other's base class.
            # If self and other don't have a common base, we don't know how to combine them.
            from fipy.variables.constant import _Constant
            if isinstance(self, other._getArithmeticBaseClass()) or isinstance(other, _Constant):
                return self._getArithmeticBaseClass()
            else:
                return None
        else:
            # If self and other have un-broadcastable shapes, we don't know how to combine them.
            return None

    def _OperatorVariableClass(self, baseClass=None):
        from fipy.variables import operatorVariable
        
        baseClass = baseClass or self._variableClass
        return operatorVariable._OperatorVariableClass(baseClass=baseClass)
            
    def _UnaryOperatorVariable(self, op, operatorClass=None, opShape=None, canInline=True, unit=None):
        """
        Check that unit works for unOp

            >>> (-Variable(value="1 m")).unit
            <PhysicalUnit m>
            
        """
        operatorClass = operatorClass or self._OperatorVariableClass()
        from fipy.variables import unaryOperatorVariable
        unOp = unaryOperatorVariable._UnaryOperatorVariable(operatorClass)
        
        # If the caller has not specified a shape for the result, determine the 
        # shape from the base class or from the inputs
        if opShape is None:
            opShape = self.shape
        
        if opShape is None:
            return NotImplemented

        if not self.unit.isDimensionless():
            canInline = False

        return unOp(op=op, var=[self], opShape=opShape, canInline=canInline, unit=unit, 
                    inlineComment=inline._operatorVariableComment(canInline=canInline))

    def _shapeClassAndOther(self, opShape, operatorClass, other):
        """
        Determine the shape of the result, the base class of the result, and (if
        necessary) a modified form of `other` that is suitable for the
        operation.
        """
        # If the caller has not specified a base class for the binop, 
        # check if the member Variables know what type of Variable should
        # result from the operation.
        baseClass = operatorClass or self._getArithmeticBaseClass(other)
    
        # If the caller has not specified a shape for the result, determine the 
        # shape from the base class or from the inputs
        if opShape is None:
            opShape = self._broadcastShape(other)

        return (opShape, baseClass, other)
        
    def _BinaryOperatorVariable(self, op, other, operatorClass=None, opShape=None, canInline=True, unit=None):
        """
        :Parameters:
          - `op`: the operator function to apply (takes two arguments for `self` and `other`)
          - `other`: the quantity to be operated with
          - `operatorClass`: the `Variable` class that the binary operator should inherit from 
          - `opShape`: the shape that should result from the operation
        """
        if not isinstance(other, Variable):
            from fipy.variables.constant import _Constant
            other = _Constant(value=other)

        opShape, baseClass, other = self._shapeClassAndOther(opShape, operatorClass, other)

        if opShape is None or baseClass is None:
            return NotImplemented
        
        for v in [self, other]:
            if not v.unit.isDimensionless() or len(v.shape) > 3:
                canInline = False
                
        # obtain a general operator class with the desired base class
        operatorClass = operatorClass or self._OperatorVariableClass(baseClass)
        from fipy.variables import binaryOperatorVariable
        binOp = binaryOperatorVariable._BinaryOperatorVariable(operatorClass)
        
        return binOp(op=op, var=[self, other], opShape=opShape, canInline=canInline, unit=unit, 
                     inlineComment=inline._operatorVariableComment(canInline=canInline))
    
    def __add__(self, other):
        from fipy.terms.term import Term
        if isinstance(other, Term):
            return other + self
        else:
            return self._BinaryOperatorVariable(lambda a,b: a+b, other)
        
    __radd__ = __add__

    def __sub__(self, other):
        from fipy.terms.term import Term
        if isinstance(other, Term):
            return -other + self
        else:
            return self._BinaryOperatorVariable(lambda a,b: a-b, other)
        
    def __rsub__(self, other):
        return self._BinaryOperatorVariable(lambda a,b: b-a, other)
            
    def __mul__(self, other):
        from fipy.terms.term import Term
        if isinstance(other, Term):
            return other * self
        else:
            return self._BinaryOperatorVariable(lambda a,b: a*b, other)

    __rmul__ = __mul__
    
    def __mod__(self, other):
        return self._BinaryOperatorVariable(lambda a,b: numerix.fmod(a, b), other)
            
    def __pow__(self, other):
        return self._BinaryOperatorVariable(lambda a,b: pow(a,b), other)
            
    def __rpow__(self, other):
        return self._BinaryOperatorVariable(lambda a,b: pow(b,a), other)
            
    def __truediv__(self, other):
        return self._BinaryOperatorVariable(lambda a,b: a/b, other)
        
    __div__ = __truediv__
    
    def __rtruediv__(self, other):
        return self._BinaryOperatorVariable(lambda a,b: b/a, other)
            
    __rdiv__ = __rtruediv__
    
    def __neg__(self):
        return self._UnaryOperatorVariable(lambda a: -a)
        
    def __pos__(self):
        return self
        
    def __abs__(self):
        """

        Following test it to fix a bug with C inline string using
        abs() instead of fabs()

            >>> print abs(Variable(2.3) - Variable(1.2))
            1.1

        """
        return self._UnaryOperatorVariable(lambda a: numerix.fabs(a))

    def __invert__(self):
        """
        Returns logical "not" of the `Variable`
        
            >>> a = Variable(value=True)
            >>> print ~a
            False
        """
        return self._UnaryOperatorVariable(lambda a: ~a)

    def __lt__(self,other):
        """
        Test if a `Variable` is less than another quantity
        
            >>> a = Variable(value=3)
            >>> b = (a < 4)
            >>> b
            (Variable(value=array(3)) < 4)
            >>> b()
            1
            >>> a.value = 4
            >>> print b()
            0
            >>> print 1000000000000000000 * Variable(1) < 1.
            0
            >>> print 1000 * Variable(1) < 1.
            0


        Python automatically reverses the arguments when necessary
        
            >>> 4 > Variable(value=3)
            (Variable(value=array(3)) < 4)
        """
        return self._BinaryOperatorVariable(lambda a,b: a<b, other)

    def __le__(self,other):
        """
        Test if a `Variable` is less than or equal to another quantity
        
            >>> a = Variable(value=3)
            >>> b = (a <= 4)
            >>> b
            (Variable(value=array(3)) <= 4)
            >>> b()
            1
            >>> a.value = 4
            >>> print b()
            1
            >>> a.value = 5
            >>> print b()
            0
        """
        return self._BinaryOperatorVariable(lambda a,b: a<=b, other)
        
    def __eq__(self,other):
        """
        Test if a `Variable` is equal to another quantity
        
            >>> a = Variable(value=3)
            >>> b = (a == 4)
            >>> b
            (Variable(value=array(3)) == 4)
            >>> b()
            0
        """
        return self._BinaryOperatorVariable(lambda a,b: a==b, other)
        
    __hash__ = object.__hash__
    
    def __ne__(self,other):
        """
        Test if a `Variable` is not equal to another quantity
        
            >>> a = Variable(value=3)
            >>> b = (a != 4)
            >>> b
            (Variable(value=array(3)) != 4)
            >>> b()
            1
        """
        return self._BinaryOperatorVariable(lambda a,b: a!=b, other)
        
    def __gt__(self,other):
        """
        Test if a `Variable` is greater than another quantity
        
            >>> a = Variable(value=3)
            >>> b = (a > 4)
            >>> b
            (Variable(value=array(3)) > 4)
            >>> print b()
            0
            >>> a.value = 5
            >>> print b()
            1
        """
        return self._BinaryOperatorVariable(lambda a,b: a>b, other)
        
    def __ge__(self,other):
        """
        Test if a `Variable` is greater than or equal to another quantity
        
            >>> a = Variable(value=3)
            >>> b = (a >= 4)
            >>> b
            (Variable(value=array(3)) >= 4)
            >>> b()
            0
            >>> a.value = 4
            >>> print b()
            1
            >>> a.value = 5
            >>> print b()
            1
        """
        return self._BinaryOperatorVariable(lambda a,b: a>=b, other)

    def __and__(self, other):
        """
        This test case has been added due to a weird bug that was appearing.

        >>> a = Variable(value=(0, 0, 1, 1))
        >>> b = Variable(value=(0, 1, 0, 1))
        >>> print numerix.equal((a == 0) & (b == 1), [False,  True, False, False]).all()
        True
        >>> print a & b
        [0 0 0 1]
        >>> from fipy.meshes import Grid1D
        >>> mesh = Grid1D(nx=4)
        >>> from fipy.variables.cellVariable import CellVariable
        >>> a = CellVariable(value=(0, 0, 1, 1), mesh=mesh)
        >>> b = CellVariable(value=(0, 1, 0, 1), mesh=mesh)
        >>> print numerix.allequal((a == 0) & (b == 1), [False,  True, False, False])
        True
        >>> print a & b
        [0 0 0 1]
        """
        return self._BinaryOperatorVariable(lambda a,b: a & b, other, canInline=False)

    def __or__(self, other):
        """
        This test case has been added due to a weird bug that was appearing.

        >>> a = Variable(value=(0, 0, 1, 1))
        >>> b = Variable(value=(0, 1, 0, 1))
        >>> print numerix.equal((a == 0) | (b == 1), [True,  True, False, True]).all()
        True
        >>> print a | b
        [0 1 1 1]
        >>> from fipy.meshes import Grid1D
        >>> mesh = Grid1D(nx=4)
        >>> from fipy.variables.cellVariable import CellVariable
        >>> a = CellVariable(value=(0, 0, 1, 1), mesh=mesh)
        >>> b = CellVariable(value=(0, 1, 0, 1), mesh=mesh)
        >>> print numerix.allequal((a == 0) | (b == 1), [True,  True, False, True])
        True
        >>> print a | b
        [0 1 1 1]
        """
        return self._BinaryOperatorVariable(lambda a,b: a | b, other, canInline=False)

    def __iter__(self):
        return iter(self.value)
        
    def __len__(self):
        return len(self.value)
        
    def __float__(self):
        return float(self.value)
        
    def __int__(self):
        return int(self.value)
        
    def __bool__(self):
        """
            >>> print bool(Variable(value=0))
            0
            >>> print bool(Variable(value=(0, 0, 1, 1)))
            Traceback (most recent call last):
                ...
            ValueError: The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()
        """
        return bool(self.value)
    
    def any(self, axis=None):
        """
            >>> print Variable(value=0).any()
            0
            >>> print Variable(value=(0, 0, 1, 1)).any()
            1
        """
        operatorClass = Variable._OperatorVariableClass(self, baseClass=Variable)
        return self._UnaryOperatorVariable(lambda a: a.any(axis=axis),
                                           operatorClass=operatorClass,
                                           opShape=(),
                                           canInline=False)

    def all(self, axis=None):
        """
            >>> print Variable(value=(0, 0, 1, 1)).all()
            0
            >>> print Variable(value=(1, 1, 1, 1)).all()
            1
        """
        operatorClass = Variable._OperatorVariableClass(self, baseClass=Variable)
        return self._UnaryOperatorVariable(lambda a: a.all(axis=axis),
                                           operatorClass=operatorClass,
                                           opShape=(),
                                           canInline=False)

    @mathMethodDeprecated
    def arccos(self):
        return self._UnaryOperatorVariable(lambda a: numerix.arccos(a))

    @mathMethodDeprecated
    def arccosh(self):
        return self._UnaryOperatorVariable(lambda a: numerix.arccosh(a))

    @mathMethodDeprecated
    def arcsin(self):
        return self._UnaryOperatorVariable(lambda a: numerix.arcsin(a))

    @mathMethodDeprecated
    def arcsinh(self):
        return self._UnaryOperatorVariable(lambda a: numerix.arcsinh(a))

    @mathMethodDeprecated
    def sqrt(self):
        """
        
            >>> from fipy.meshes import Grid1D
            >>> mesh= Grid1D(nx=3)

            >>> from fipy.variables.cellVariable import CellVariable
            >>> var = CellVariable(mesh=mesh, value=((0., 2., 3.),), rank=1)
            >>> print (var.dot(var)).sqrt()
            [ 0.  2.  3.]
            
        """
        return self._UnaryOperatorVariable(lambda a: numerix.sqrt(a))
        
    @mathMethodDeprecated
    def tan(self):
        return self._UnaryOperatorVariable(lambda a: numerix.tan(a))

    @mathMethodDeprecated
    def tanh(self):
        return self._UnaryOperatorVariable(lambda a: numerix.tanh(a))

    @mathMethodDeprecated
    def arctan(self):
        return self._UnaryOperatorVariable(lambda a: numerix.arctan(a))

    @mathMethodDeprecated
    def arctanh(self):
        return self._UnaryOperatorVariable(lambda a: numerix.arctanh(a))
            
    @mathMethodDeprecated
    def exp(self):
        return self._UnaryOperatorVariable(lambda a: numerix.exp(a))

    @mathMethodDeprecated
    def log(self):
        return self._UnaryOperatorVariable(lambda a: numerix.log(a))

    @mathMethodDeprecated
    def log10(self):
        return self._UnaryOperatorVariable(lambda a: numerix.log10(a))

    @mathMethodDeprecated
    def sin(self):
        return self._UnaryOperatorVariable(lambda a: numerix.sin(a))
                
    @mathMethodDeprecated
    def sinh(self):
        return self._UnaryOperatorVariable(lambda a: numerix.sinh(a))

    @mathMethodDeprecated
    def cos(self):
        return self._UnaryOperatorVariable(lambda a: numerix.cos(a))
        
    @mathMethodDeprecated
    def cosh(self):
        return self._UnaryOperatorVariable(lambda a: numerix.cosh(a))

    @mathMethodDeprecated
    def floor(self):
        return self._UnaryOperatorVariable(lambda a: numerix.floor(a))

    @mathMethodDeprecated
    def ceil(self):
        return self._UnaryOperatorVariable(lambda a: numerix.ceil(a))

    @mathMethodDeprecated
    def sign(self):
        return self._UnaryOperatorVariable(lambda a: numerix.sign(a), canInline=False)
        
    @mathMethodDeprecated
    def conjugate(self):
        return self._UnaryOperatorVariable(lambda a: numerix.conjugate(a), canInline=False)

    @mathMethodDeprecated
    def arctan2(self, other):
        return self._BinaryOperatorVariable(lambda a,b: numerix.arctan2(a,b), other)
        
    def dot(self, other, opShape=None, operatorClass=None, axis=0):
        if not isinstance(other, Variable):
            from fipy.variables.constant import _Constant
            other = _Constant(value=other)
        if opShape is None:
            opShape = self._broadcastShape(other)
        return self._BinaryOperatorVariable(lambda a,b: numerix.dot(a,b, axis=axis), 
                                            other, 
                                            opShape=opShape[:axis]+opShape[axis+1:],
                                            operatorClass=operatorClass,
                                            canInline=False)
        
    @mathMethodDeprecated
    def reshape(self, shape):
        return self._BinaryOperatorVariable(lambda a,b: numerix.reshape(a,b), shape, opShape=shape, canInline=False)

    def ravel(self):
        return self.value.ravel()
        
    def _axisClass(self, axis):
        return self._OperatorVariableClass()

    def _axisOperator(self, opname, op, axis=None):
        if not hasattr(self, opname):
            setattr(self, opname, {})
            
        opdict = getattr(self, opname)
        if axis not in opdict:
            if axis is None:
                opShape = ()
            else:
                opShape=self.shape[:axis] + self.shape[axis+1:]
                
            opdict[axis] = self._UnaryOperatorVariable(op,
                                                       operatorClass=self._axisClass(axis=axis), 
                                                       opShape=opShape,
                                                       canInline=False)
        
        return opdict[axis]

    def sum(self, axis=None):
        return self._axisOperator(opname="sumVar", 
                                  op=lambda a: numerix.sum(a, axis=axis), 
                                  axis=axis)
                                    
    def max(self, axis=None):
        return self._axisOperator(opname="maxVar", 
                                  op=lambda a: a.max(axis=axis), 
                                  axis=axis)
                                  
    def min(self, axis=None):
        return self._axisOperator(opname="minVar", 
                                  op=lambda a: a.min(axis=axis), 
                                  axis=axis)
                                  
    def _getitemClass(self, index):
        return self._OperatorVariableClass()

    def __getitem__(self, index):
        """    
        "Evaluate" the `Variable` and return the specified element
      
            >>> a = Variable(value=((3.,4.),(5.,6.)), unit="m") + "4 m"
            >>> print a[1,1]
            10.0 m

        It is an error to slice a `Variable` whose `value` is not sliceable

            >>> Variable(value=3)[2]
            Traceback (most recent call last):
                  ...
            IndexError: 0-d arrays can't be indexed

        """
        return self._UnaryOperatorVariable(lambda a: a[index], 
                                           operatorClass=self._getitemClass(index=index), 
                                           opShape=numerix._indexShape(index=index, arrayShape=self.shape),
                                           unit=self.unit,
                                           canInline=False)

    def take(self, ids, axis=0):
        return numerix.take(self.value, ids, axis)

    def allclose(self, other, rtol=1.e-5, atol=1.e-8):
        """
           >>> var = Variable((1, 1))
           >>> print var.allclose((1, 1))
           1
           >>> print var.allclose((1,))
           1

        The following test is to check that the system does not run
        out of memory.

           >>> from fipy.tools import numerix
           >>> var = Variable(numerix.ones(10000))
           >>> print var.allclose(numerix.zeros(10000, 'l'))
           False
           
        """
        operatorClass = Variable._OperatorVariableClass(self, baseClass=Variable)
        return self._BinaryOperatorVariable(lambda a,b: numerix.allclose(a, b, atol=atol, rtol=rtol), 
                                            other, 
                                            operatorClass=operatorClass,
                                            opShape=(),
                                            canInline=False)
        
    def allequal(self, other):
        operatorClass = Variable._OperatorVariableClass(self, baseClass=Variable)
        return self._BinaryOperatorVariable(lambda a,b: numerix.allequal(a,b), 
                                            other,
                                            operatorClass=operatorClass,
                                            opShape=(),
                                            canInline=False)

    @getsetDeprecated
    def getMag(self):
        return self.mag
    
    @property
    def mag(self):
        if not hasattr(self, "_mag"):
            self._mag = numerix.sqrt(self.dot(self))
            
        return self._mag
    
    def __getstate__(self):
        """
        Used internally to collect the necessary information to ``pickle`` the 
        `Variable` to persistent storage.
        """
        return {
            'value': self.value,
            'unit': self.unit,
            'array': None,
            'name': self.name,
        }
        
    def __setstate__(self, dict):
        """
        Used internally to create a new `Variable` from ``pickled`` 
        persistent storage.
        """
        
        import sys
        self._refcount = sys.getrefcount(self)

        self.__init__(**dict)
        
    def _test(self):
        """
        Inverse cosine of :math:`x`, :math:`\cos^{-1} x`
        
        >>> from fipy.tools.numerix import *

        >>> arccos(Variable(value=(0,0.5,1.0)))
        arccos(Variable(value=array([ 0. ,  0.5,  1. ])))
            
        .. attention:: 
            
           the next should really return radians, but doesn't
           
        >>> print tostring(arccos(Variable(value=(0,0.5,1.0))), precision=3)
        [ 1.571  1.047  0.   ]

        Inverse hyperbolic cosine of :math:`x`, :math:`\cosh^{-1} x`
        
        >>> arccosh(Variable(value=(1,2,3)))
        arccosh(Variable(value=array([1, 2, 3])))
        >>> print tostring(arccosh(Variable(value=(1,2,3))), precision=3)
        [ 0.     1.317  1.763]
        
        Inverse sine of :math:`x`, :math:`\sin^{-1} x`
        
        >>> arcsin(Variable(value=(0,0.5,1.0)))
        arcsin(Variable(value=array([ 0. ,  0.5,  1. ])))
            
        .. attention:: 
            
           the next should really return radians, but doesn't
           
        >>> print tostring(arcsin(Variable(value=(0,0.5,1.0))), precision=3)
        [ 0.     0.524  1.571]

        Inverse hyperbolic sine of :math:`x`, :math:`\sinh^{-1} x`

        >>> arcsinh(Variable(value=(1,2,3)))
        arcsinh(Variable(value=array([1, 2, 3])))
        >>> print tostring(arcsinh(Variable(value=(1,2,3))), precision=3)
        [ 0.881  1.444  1.818]
        
        Inverse tangent of :math:`x`, :math:`\tan^{-1} x`

        >>> arctan(Variable(value=(0,0.5,1.0)))
        arctan(Variable(value=array([ 0. ,  0.5,  1. ])))
        
        .. attention:: 
            
           the next should really return radians, but doesn't
           
        >>> print tostring(arctan(Variable(value=(0,0.5,1.0))), precision=3)
        [ 0.     0.464  0.785]

        Inverse tangent of a ratio :math:`x/y`, :math:`\tan^{-1} \frac{x}{y}`

        >>> arctan2(Variable(value=(0, 1, 2)), 2)
        (arctan2(Variable(value=array([0, 1, 2])), 2))
            
        .. attention:: 
            
           the next should really return radians, but doesn't
           
        >>> print tostring(arctan2(Variable(value=(0, 1, 2)), 2), precision=3)
        [ 0.     0.464  0.785]

        Inverse hyperbolic tangent of :math:`x`, :math:`\tanh^{-1} x`

        >>> arctanh(Variable(value=(0,0.25,0.5)))
        arctanh(Variable(value=array([ 0.  ,  0.25,  0.5 ])))
        >>> print tostring(arctanh(Variable(value=(0,0.25,0.5))), precision=3)
        [ 0.     0.255  0.549]

        Cosine of :math:`x`, :math:`\cos x`

        >>> cos(Variable(value=(0,2*pi/6,pi/2), unit="rad"))
        cos(Variable(value=PhysicalField(array([ 0.        ,  1.04719755,  1.57079633]),'rad')))
        >>> print tostring(cos(Variable(value=(0,2*pi/6,pi/2), unit="rad")), suppress_small=1)
        [ 1.   0.5  0. ]
        
        Hyperbolic cosine of :math:`x`, :math:`\cosh x`

        >>> cosh(Variable(value=(0,1,2)))
        cosh(Variable(value=array([0, 1, 2])))
        >>> print tostring(cosh(Variable(value=(0,1,2))), precision=3)
        [ 1.     1.543  3.762]
        
        Tangent of :math:`x`, :math:`\tan x`

        >>> tan(Variable(value=(0,pi/3,2*pi/3), unit="rad"))
        tan(Variable(value=PhysicalField(array([ 0.        ,  1.04719755,  2.0943951 ]),'rad')))
        >>> print tostring(tan(Variable(value=(0,pi/3,2*pi/3), unit="rad")), precision=3)
        [ 0.     1.732 -1.732]
        
        Hyperbolic tangent of :math:`x`, :math:`\tanh x`

        >>> tanh(Variable(value=(0,1,2)))
        tanh(Variable(value=array([0, 1, 2])))
        >>> print tostring(tanh(Variable(value=(0,1,2))), precision=3)
        [ 0.     0.762  0.964]
        
        Base-10 logarithm of :math:`x`, :math:`\log_{10} x`

        >>> log10(Variable(value=(0.1,1,10)))
        log10(Variable(value=array([  0.1,   1. ,  10. ])))
        >>> print log10(Variable(value=(0.1,1,10)))
        [-1.  0.  1.]
        
        Sine of :math:`x`, :math:`\sin x`

        >>> sin(Variable(value=(0,pi/6,pi/2), unit="rad"))
        sin(Variable(value=PhysicalField(array([ 0.        ,  0.52359878,  1.57079633]),'rad')))
        >>> print sin(Variable(value=(0,pi/6,pi/2), unit="rad"))
        [ 0.   0.5  1. ]
        
        Hyperbolic sine of :math:`x`, :math:`\sinh x`

        >>> sinh(Variable(value=(0,1,2)))
        sinh(Variable(value=array([0, 1, 2])))
        >>> print tostring(sinh(Variable(value=(0,1,2))), precision=3)
        [ 0.     1.175  3.627]

        Square root of :math:`x`, :math:`\sqrt{x}`

        >>> sqrt(Variable(value=(1, 2, 3), unit="m**2"))
        sqrt(Variable(value=PhysicalField(array([1, 2, 3]),'m**2')))
        >>> print tostring(sqrt(Variable(value=(1, 2, 3), unit="m**2")), precision=3)
        [ 1.     1.414  1.732] m
        
        The largest integer :math:`\le x`, :math:`\lfloor x \rfloor`

        >>> floor(Variable(value=(-1.5,2,2.5), unit="m**2"))
        floor(Variable(value=PhysicalField(array([-1.5,  2. ,  2.5]),'m**2')))
        >>> print floor(Variable(value=(-1.5,2,2.5), unit="m**2"))
        [-2.  2.  2.] m**2
        
        The largest integer :math:`\ge x`, :math:`\lceil x \rceil`
           
        >>> ceil(Variable(value=(-1.5,2,2.5), unit="m**2"))
        ceil(Variable(value=PhysicalField(array([-1.5,  2. ,  2.5]),'m**2')))
        >>> print ceil(Variable(value=(-1.5,2,2.5), unit="m**2"))
        [-1.  2.  3.] m**2
        
        Natural logarithm of :math:`x`, :math:`\ln x \equiv \log_e x`

        >>> log(Variable(value=(0.1,1,10)))
        log(Variable(value=array([  0.1,   1. ,  10. ])))
        >>> print tostring(log(Variable(value=(0.1,1,10))), precision=3)
        [-2.303  0.     2.303]
        
        Complex conjugate of :math:`z = x + i y`, :math:`z^\star = x - i y`
           
        >>> var = conjugate(Variable(value=(3 + 4j, -2j, 10), unit="ohm"))
        >>> print var.unit
        <PhysicalUnit ohm>
        >>> print allclose(var.numericValue, (3 - 4j, 2j, 10))
        1
        """
        pass
        

def _test(): 
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()
    
if __name__ == "__main__": 
    _test() 



#!/usr/bin/env python

## 
 # -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "pysparseSolver.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

__all__ = []

from fipy.solvers.solver import Solver
from fipy.matrices.pysparseMatrix import _PysparseMeshMatrix

from fipy.tools.decorators import getsetDeprecated
 
class _PysparseMatrixSolver(Solver):

    """
    A class consolidating methods for solver packages which use
    `_PysparseMeshMatrix` for their matrix class.

    Subclasses have a `_solve_` method, which is called by `_solve`. Typically,
    `_solve_` returns the new value of `self.var` to `_solve` and solve sets the
    var accordingly.

    A solution function `solveFnc`, usually of the form `solve(A, x, b)`, is
    implemented in most leaf-node child classes.

    .. attention:: This class is abstract. Always create one of its subclasses.
    """

    solveFnc = None

    @getsetDeprecated
    def _getMatrixClass(self):
        return self._matrixClass

    @property
    def _matrixClass(self):
        return _PysparseMeshMatrix
         
    def _solve(self):
        """
        Call `_solve_` for the new value of `self.var`.

        In certain cases, `_solve_` won't return anything, e.g. 
        `fipy.solvers.pysparse.linearLUSolver`. In these cases, we preserve the
        value of `self.var.numericValue`.
        """

        if self.var.mesh.communicator.Nproc > 1:
            raise Exception("%ss cannot be used with multiple processors" \
                            % self.__class__)
        
        array = self.var.numericValue
        newArr = self._solve_(self.matrix, array, self.RHSvector)

        if newArr is not None:
            array = newArr

        factor = self.var.unit.factor

        if factor != 1:
            array /= self.var.unit.factor

        self.var[:] = array  



from fipy.solvers.trilinos.preconditioners.multilevelDDPreconditioner import *
from fipy.solvers.trilinos.preconditioners.multilevelSAPreconditioner import *
from fipy.solvers.trilinos.preconditioners.multilevelDDMLPreconditioner import *
from fipy.solvers.trilinos.preconditioners.multilevelNSSAPreconditioner import *
from fipy.solvers.trilinos.preconditioners.jacobiPreconditioner import *
from fipy.solvers.trilinos.preconditioners.icPreconditioner import *
from fipy.solvers.trilinos.preconditioners.domDecompPreconditioner import *
from fipy.solvers.trilinos.preconditioners.multilevelSGSPreconditioner import *
from fipy.solvers.trilinos.preconditioners.multilevelSolverSmootherPreconditioner import *

__all__ = []
__all__.extend(multilevelDDPreconditioner.__all__)
__all__.extend(multilevelSAPreconditioner.__all__)
__all__.extend(multilevelDDMLPreconditioner.__all__)
__all__.extend(multilevelNSSAPreconditioner.__all__)
__all__.extend(jacobiPreconditioner.__all__)
__all__.extend(icPreconditioner.__all__)
__all__.extend(domDecompPreconditioner.__all__)
__all__.extend(multilevelSGSPreconditioner.__all__)
__all__.extend(multilevelSolverSmootherPreconditioner.__all__)



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - a finite volume PDE solver in Python
 # 
 #  FILE: "gaussianNoiseVariable.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This document was prepared at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this document is not subject to copyright
 # protection and is in the public domain.  langevinNoiseTerm.py
 # is an experimental work.  NIST assumes no responsibility whatsoever
 # for its use by other parties, and makes no guarantees, expressed
 # or implied, about its quality, reliability, or any other characteristic.
 # We would appreciate acknowledgement if the document is used.
 # 
 # This document can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution
 #  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

from fipy.tools.numerix import random, sqrt
from fipy.variables.noiseVariable import NoiseVariable

__all__ = ["GaussianNoiseVariable"]

class GaussianNoiseVariable(NoiseVariable):
    r"""
    
    Represents a normal (Gaussian) distribution of random numbers with 
    mean :math:`\mu` and variance
    :math:`\langle \eta(\vec{r}) \eta(\vec{r}\,') \rangle = \sigma^2`,
    which has the probability distribution
    
    .. math::
        
       \frac{1}{\sigma\sqrt{2\pi}} \exp -\frac{(x - \mu)^2}{2\sigma^2}

    For example, the variance of thermal noise that is uncorrelated in space and
    time is often expressed as
    
    .. math::
    
       \left\langle
           \eta(\vec{r}, t) \eta(\vec{r}\,', t')
       \right\rangle = 
       M k_B T \delta(\vec{r} - \vec{r}\,')\delta(t - t') 
       
    which can be obtained with::
        
        sigmaSqrd = Mobility * kBoltzmann * Temperature / (mesh.cellVolumes * timeStep)
        GaussianNoiseVariable(mesh = mesh, variance = sigmaSqrd)

    .. note:: 
        
       If the time step will change as the simulation progresses, either through
       use of an adaptive iterator or by making manual changes at different
       stages, remember to declare `timeStep` as a `Variable` and to change its
       value with its `setValue()` method.
    
    >>> import sys
    >>> from fipy.tools.numerix import *

    >>> mean = 0.
    >>> variance = 4.

    Seed the random module for the sake of deterministic test results.

    >>> from fipy import numerix
    >>> numerix.random.seed(3)

    We generate noise on a non-uniform cartesian mesh with cell dimensions of
    :math:`x^2` and :math:`y^3`.
           
    >>> from fipy.meshes import Grid2D
    >>> mesh = Grid2D(dx = arange(0.1, 5., 0.1)**2, dy = arange(0.1, 3., 0.1)**3)
    >>> from fipy.variables.cellVariable import CellVariable
    >>> volumes = CellVariable(mesh=mesh,value=mesh.cellVolumes)
    >>> noise = GaussianNoiseVariable(mesh = mesh, mean = mean, 
    ...                               variance = variance / volumes)
           
    We histogram the root-volume-weighted noise distribution
    
    >>> from fipy.variables.histogramVariable import HistogramVariable
    >>> histogram = HistogramVariable(distribution = noise * sqrt(volumes), 
    ...                               dx = 0.1, nx = 600, offset = -30)
           
    and compare to a Gaussian distribution
    
    >>> gauss = CellVariable(mesh = histogram.mesh)
    >>> x = histogram.mesh.cellCenters[0]
    >>> gauss.value = ((1/(sqrt(variance * 2 * pi))) * exp(-(x - mean)**2 / (2 * variance)))
    
    >>> if __name__ == '__main__':
    ...     from fipy.viewers import Viewer
    ...     viewer = Viewer(vars=noise, 
    ...                     datamin=-5, datamax=5)
    ...     histoplot = Viewer(vars=(histogram, gauss))
    
    >>> for i in range(10):
    ...     noise.scramble()
    ...     if __name__ == '__main__':
    ...         viewer.plot()
    ...         histoplot.plot()

    >>> print abs(noise.faceGrad.divergence.cellVolumeAverage) < 5e-15
    1

    Note that the noise exhibits larger amplitude in the small cells than in the large ones

    .. image:: fipy/variables/gaussian.*
      :scale: 25
      :align: center
      :alt: random values with a gaussian distribution

    but that the root-volume-weighted histogram is Gaussian.

    .. image:: fipy/variables/gauss-histogram.*
      :scale: 25
      :align: center
      :alt: histogram of random values with a gaussian distribution

    """
    def __init__(self, mesh, name = '', mean = 0., variance = 1., hasOld = 0):
        """
        :Parameters:
            - `mesh`: The mesh on which to define the noise.
            - `mean`: The mean of the noise distrubution, :math:`\mu`.
            - `variance`: The variance of the noise distribution, :math:`\sigma^2`.
        """
        self.mean = mean
        self.variance = variance
        NoiseVariable.__init__(self, mesh = mesh, name = name, hasOld = hasOld)

    def parallelRandom(self):

        if hasattr(self.variance, 'globalValue'):
            variance = self.variance.globalValue
        else:
            variance = self.variance

        if self.mesh.communicator.procID == 0:
            return random.normal(self.mean, sqrt(variance),
                                 size = [self.mesh.globalNumberOfCells])
        else:
            return None

def _test(): 
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()
    
if __name__ == "__main__": 
    _test() 




#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "gistViewer.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution
 #  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

from fipy.viewers.gistViewer.gistViewer import _GistViewer
from fipy.variables.cellVariable import CellVariable
from fipy.variables.faceVariable import FaceVariable
from fipy.tools import numerix
from fipy.tools.decorators import deprecateGist
__all__ = ["GistVectorViewer"]

@deprecateGist
class GistVectorViewer(_GistViewer):
    """Displays a vector plot of a 2D rank-1 `CellVariable` or
    `FaceVariable` object using gist.
    """

    __doc__ += _GistViewer._test2Dvector(viewer="GistVectorViewer")
    __doc__ += _GistViewer._test2DvectorIrregular(viewer="GistVectorViewer")
    
    def __init__(self, vars, title=None, limits={}, **kwlimits):
        """Creates a `GistVectorViewer`.
        
        :Parameters:
          vars
            a rank-1 `CellVariable` or `FaceVariable` object.
          title
            displayed at the top of the `Viewer` window
          limits : dict
            a (deprecated) alternative to limit keyword arguments
          xmin, xmax, ymin, ymax, datamin, datamax
            displayed range of data. Any limit set to 
            a (default) value of `None` will autoscale.
        """
        kwlimits.update(limits)
	_GistViewer.__init__(self, vars=vars, title=title, **kwlimits)
        
    def _getSuitableVars(self, vars):
        vars = [var for var in _GistViewer._getSuitableVars(self, vars) \
          if (var.mesh.dim == 2 \
              and (isinstance(var, FaceVariable) \
                   or isinstance(var, CellVariable)) and var.rank == 1)]
        if len(vars) == 0:
            from fipy.viewers import MeshDimensionError
            raise MeshDimensionError("Can only plot 2D vector data")
        # this viewer can only display one variable
        return [vars[0]]
        
    def plot(self, filename=None):
        import gist

        gist.window(self.id, wait = 1)
        gist.pltitle(self.title)
        gist.animate(1)
        
        var = self.vars[0]
        
        if isinstance(var, FaceVariable):
            x, y = var.mesh.faceCenters
        elif isinstance(var, CellVariable):
            x, y = var.mesh.cellCenters
        
        gist.plmesh(numerix.array([y, y]), numerix.array([x, y]))

        vx = numerix.array(var[0])
        vy = numerix.array(var[1])
        
        maxVec = var.mag.max().value
        maxGrid = var.mesh._cellDistances.max()
        
        gist.plv(numerix.array([vy,vy]), numerix.array([vx,vx]), scale=maxGrid / maxVec * 3, hollow=1, aspect=0.25) #,scale=0.002)
        
        if filename is not None:
            
            gist.hcp_file(filename)
            gist.hcp()

        gist.fma()
        
    def getArray(self):
        pass
        
if __name__ == "__main__": 
    import fipy.tests.doctestPlus
    fipy.tests.doctestPlus.execButNoTest()



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "gist1DViewer.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution
 #  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

from fipy.tools import numerix
from fipy.viewers.gistViewer.gistViewer import _GistViewer
from fipy.tools.decorators import deprecateGist
__all__ = ["Gist1DViewer"]

@deprecateGist
class Gist1DViewer(_GistViewer):
    """Displays a y vs. x plot of one or more 1D `CellVariable` objects.
    """
    
    __doc__ += _GistViewer._test1D(viewer="Gist1DViewer")

    def __init__(self, vars, title=None, xlog=0, ylog=0, style="work.gs", limits={}, **kwlimits):
        """
        Creates a `Gist1DViewer`.
        
        :Parameters:
          vars
            a `CellVariable` or tuple of `CellVariable` objects to plot
          title
            displayed at the top of the `Viewer` window
          xlog
            log scaling of x axis if `True`
          ylog
            log scaling of y axis if `True`
          stye
            the Gist stylefile to use.
          limits : dict
            a (deprecated) alternative to limit keyword arguments
          xmin, xmax, datamin, datamax
            displayed range of data. Any limit set to 
            a (default) value of `None` will autoscale.
            (*ymin* and *ymax* are synonyms for *datamin* and *datamax*).
        """
        kwlimits.update(limits)
        _GistViewer.__init__(self, vars=vars, title=title, **kwlimits)
        
        self.xlog = xlog
        self.ylog = ylog
        self.style = style
        
    def _getSuitableVars(self, vars):
        from fipy.variables.cellVariable import CellVariable
        vars = [var for var in _GistViewer._getSuitableVars(self, vars) \
          if (var.mesh.dim == 1 and isinstance(var, CellVariable))]
        if len(vars) > 1:
            vars = [var for var in vars if var.mesh is vars[0].mesh]
        if len(vars) == 0:
            from fipy.viewers import MeshDimensionError
            raise MeshDimensionError("Can only plot 1D data")
        return vars
        
    def _getArrays(self):
        arrays = []
        
        for var in self.vars:
            arrays.append((numerix.array(var), numerix.array(var.mesh.cellCenters[0])))
            
        return arrays
        
    def _plotArrays(self):
        import gist
        
        for array in self._getArrays():
            gist.plg(*array)
            
        gist.logxy(self.xlog, self.ylog)

    def plot(self, filename = None):
        import gist

        gist.window(self.id, wait = 1, style = self.style)
        gist.pltitle(self.title)
        gist.animate(1)

        if self.limits != None:
            gist.limits(self._getLimit('xmin'), 
                        self._getLimit('xmax'), 
                        self._getLimit(('datamin', 'ymin')), 
                        self._getLimit(('datamax', 'ymax')))
            
        self._plotArrays()
            
        _GistViewer.plot(self, filename = filename)

if __name__ == "__main__": 
    import fipy.tests.doctestPlus
    fipy.tests.doctestPlus.execButNoTest()



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "linearLUSolver.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

import os

from scipy.sparse.linalg import splu

from fipy.solvers.scipy.scipySolver import _ScipySolver
from fipy.tools import numerix

__all__ = ["LinearLUSolver"]

class LinearLUSolver(_ScipySolver):
    """
    The `LinearLUSolver` solves a linear system of equations using
    LU-factorisation.  The `LinearLUSolver` is a wrapper class for the
    the Scipy `scipy.sparse.linalg.splu` moduleq.
    """
    
    def _solve_(self, L, x, b):
        diag = L.takeDiagonal()
        maxdiag = max(numerix.absolute(diag))

        L = L * (1 / maxdiag)
        b = b * (1 / maxdiag)

        LU = splu(L.matrix.asformat("csc"), diag_pivot_thresh=1.,
                                            drop_tol=0.,
                                            relax=1,
                                            panel_size=10,
                                            permc_spec=3)

        error0 = numerix.sqrt(numerix.sum((L * x - b)**2))

        for iteration in range(min(self.iterations, 10)):
            errorVector = L * x - b

            if (numerix.sqrt(numerix.sum(errorVector**2)) / error0)  <= self.tolerance:
                break

            xError = LU.solve(errorVector)
            x[:] = x - xError
            
        if 'FIPY_VERBOSE_SOLVER' in os.environ:
            from fipy.tools.debug import PRINT        
            PRINT('iterations: %d / %d' % (iteration+1, self.iterations))
            PRINT('residual:', numerix.sqrt(numerix.sum(errorVector**2)))

        return x



#!/usr/bin/env python

## 
 # -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "scipyKrylovSolver.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #  Author: James O'Beirne <james.obeirne@gmail.com>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

__all__ = []

import os

from fipy.solvers.scipy.scipySolver import _ScipySolver

class _ScipyKrylovSolver(_ScipySolver):
    """
    The base `ScipyKrylovSolver` class.
    
    .. attention:: This class is abstract. Always create one of its subclasses.
    """
    
    def _solve_(self, L, x, b):
        A = L.matrix
        if self.preconditioner is None:
            M = None
        else:
            M = self.preconditioner._applyToMatrix(A)
            
        x, info = self.solveFnc(A, b, x, 
                                tol=self.tolerance,
                                maxiter=self.iterations,
                                M=M)

        if 'FIPY_VERBOSE_SOLVER' in os.environ:
            if info < 0:
                PRINT('failure', self._warningList[info].__class__.__name__)

        return x




#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - a finite volume PDE solver in Python
 # 
 #  FILE: "offsetSparseMatrix.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This document was prepared at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this document is not subject to copyright
 # protection and is in the public domain.  summationTerm.py
 # is an experimental work.  NIST assumes no responsibility whatsoever
 # for its use by other parties, and makes no guarantees, expressed
 # or implied, about its quality, reliability, or any other characteristic.
 # We would appreciate acknowledgement if the document is used.
 # 
 # This document can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

from fipy.tools import numerix

__all__ = ["OffsetSparseMatrix"]

def OffsetSparseMatrix(SparseMatrix, numberOfVariables, numberOfEquations):
    """
    Used in binary terms. equationIndex and varIndex need to be set statically before instantiation.
    """
    
    class OffsetSparseMatrixClass(SparseMatrix):
        equationIndex = 0
        varIndex = 0
        
        def __init__(self, mesh, bandwidth=0, sizeHint=None, 
                     numberOfVariables=numberOfVariables, numberOfEquations=numberOfEquations):
            SparseMatrix.__init__(self, mesh=mesh, bandwidth=bandwidth, sizeHint=sizeHint, 
                                  numberOfVariables=numberOfVariables, numberOfEquations=numberOfEquations) 

        def put(self, vector, id1, id2):
            SparseMatrix.put(self, vector, id1 + self.mesh.numberOfCells * self.equationIndex, id2 + self.mesh.numberOfCells * self.varIndex)

        def addAt(self, vector, id1, id2):
            SparseMatrix.addAt(self, vector, id1 + self.mesh.numberOfCells * self.equationIndex, id2 + self.mesh.numberOfCells * self.varIndex)

        def addAtDiagonal(self, vector):
            if type(vector) in [type(1), type(1.)]:
                tmp = numerix.zeros((self.mesh.numberOfCells,), 'd')
                tmp[:] = vector
                SparseMatrix.addAtDiagonal(self, tmp)
            else:
                SparseMatrix.addAtDiagonal(self, vector)

    return OffsetSparseMatrixClass



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "gnuplot1DViewer.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution
 #  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

from fipy.meshes import Grid2D
from fipy.variables.faceVariable import FaceVariable
from fipy.tools import numerix

from fipy.viewers.gnuplotViewer.gnuplotViewer import _GnuplotViewer
from fipy.tools.decorators import deprecateGnuplot

__all__ = ["Gnuplot2DViewer"]

@deprecateGnuplot
class Gnuplot2DViewer(_GnuplotViewer):
    """
    Displays a contour plot of a 2D `CellVariable` object.    
       
    The `Gnuplot2DViewer` plots a 2D `CellVariable` using a front end
    python wrapper available to download (Gnuplot.py_).

    .. _Gnuplot.py: http://gnuplot-py.sourceforge.net/
    """
    
    __doc__ += _GnuplotViewer._test2D(viewer="Gnuplot2DViewer")
    __doc__ += _GnuplotViewer._test2Dirregular(viewer="Gnuplot2DViewer")
    
    __doc__ += """
    Different style script demos_ are available at the Gnuplot_ site.

    .. _Gnuplot: http://gnuplot.sourceforge.net/
    .. _demos: http://gnuplot.sourceforge.net/demo/

    .. note::
    
        `Gnuplot2DViewer` requires Gnuplot_ version 4.0.

    """
    def __init__(self, vars, title = None, limits={}, **kwlimits):
        """Creates a `Gnuplot2DViewer`.

        :Parameters:
          vars
            a `CellVariable` object.
          title
            displayed at the top of the `Viewer` window
          limits : dict
            a (deprecated) alternative to limit keyword arguments
          xmin, xmax, ymin, ymax, datamin, datamax
            displayed range of data. Any limit set to 
            a (default) value of `None` will autoscale.
        """
        kwlimits.update(limits)
        _GnuplotViewer.__init__(self, vars=vars, title=title, **kwlimits)
        
        if len(self.vars) != 1:
            raise IndexError("A 2D Gnuplot viewer can only display one Variable")
            
    def _plot(self):

        self.g('set cbrange [' + self._getLimit(('datamin', 'zmin'))  + ':' + self._getLimit(('datamax', 'zmax')) + ']')
        self.g('set view map')
        self.g('set style data pm3d')
        self.g('set pm3d at st solid')
        mesh = self.vars[0].mesh

        if self.vars[0]._variableClass is FaceVariable:
            x, y = mesh.faceCenters
            if isinstance(mesh, Grid2D.__class__):
                nx, ny = mesh.shape
            else:
                N = int(numerix.sqrt(mesh.numberOfCells))
                nx, ny = N, N
        else:
            x, y = mesh.cellCenters
            N = int(numerix.sqrt(mesh.numberOfFaces))
            nx, ny = N, N

        self.g('set dgrid3d %i, %i, 2' % (ny, nx))

        import Gnuplot
        data = Gnuplot.Data(numerix.array(x), numerix.array(y),
                            self.vars[0].value)

        self.g.splot(data)

if __name__ == "__main__": 
    import fipy.tests.doctestPlus
    fipy.tests.doctestPlus.execButNoTest()



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "matplotlib2DViewer.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution
 #  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##
 
__docformat__ = 'restructuredtext'

from fipy.tools.decorators import getsetDeprecated

from fipy.viewers.matplotlibViewer.matplotlib2DViewer import AbstractMatplotlib2DViewer

__all__ = ["Matplotlib2DGridViewer"]

class Matplotlib2DGridViewer(AbstractMatplotlib2DViewer):
    """
    Displays an image plot of a 2D `CellVariable` object using Matplotlib_.

    .. _Matplotlib: http://matplotlib.sourceforge.net/
    """
    
    __doc__ += AbstractMatplotlib2DViewer._test2D(viewer="Matplotlib2DGridViewer")

    def __init__(self, vars, title=None, limits={}, cmap=None, colorbar='vertical', axes=None, figaspect='auto', **kwlimits):
        """
        Creates a `Matplotlib2DGridViewer`.
        
        :Parameters:
          vars
            A `CellVariable` object.
          title
            displayed at the top of the `Viewer` window
          limits : dict
            a (deprecated) alternative to limit keyword arguments
          cmap
            The colormap. Defaults to `matplotlib.cm.jet`
          xmin, xmax, ymin, ymax, datamin, datamax
            displayed range of data. Any limit set to 
            a (default) value of `None` will autoscale.
          colorbar
            plot a colorbar in specified orientation if not `None`
          axes
            if not `None`, `vars` will be plotted into this Matplotlib `Axes` object
          figaspect
            desired aspect ratio of figure. If arg is a number, use that aspect
            ratio. If arg is 'auto', the aspect ratio will be determined from
            the Variable's mesh.
        """
        kwlimits.update(limits)
        AbstractMatplotlib2DViewer.__init__(self, vars=vars, title=title, 
                                            cmap=cmap, colorbar=colorbar, axes=axes, figaspect=figaspect, 
                                            **kwlimits)

        self.image = self.axes.imshow(self._data,
                                      extent=(self._getLimit('xmin'), self._getLimit('xmax'), 
                                              self._getLimit('ymin'), self._getLimit('ymax')),
                                      vmin=0, vmax=1,
                                      cmap=self.cmap)
                   
        if title is None:                          
            self.axes.set_title(self.vars[0].name)

    def _getLimit(self, key, default=None):
        limit = AbstractMatplotlib2DViewer._getLimit(self, key, default=default)
        if limit is None:
            X, Y = self.vars[0].mesh.faceCenters
            if 'xmin' in key:
                limit = float(min(X))
            elif 'ymin' in key:
                limit = float(min(Y))
            if 'xmax' in key:
                limit = float(max(X))
            elif 'ymax' in key:
                limit = float(max(Y))
        return limit
        
    def _getSuitableVars(self, vars):
        from fipy.meshes.uniformGrid2D import UniformGrid2D
        from fipy.variables.cellVariable import CellVariable
        vars = [var for var in AbstractMatplotlib2DViewer._getSuitableVars(self, vars) \
          if (isinstance(var.mesh, UniformGrid2D) and isinstance(var, CellVariable)
              and var.rank == 0)]
        if len(vars) == 0:
            from fipy.viewers import MeshDimensionError
            raise MeshDimensionError("Matplotlib2DGridViewer can only display a rank-0 CellVariable with a UniformGrid2D mesh")
        # this viewer can only display one variable
        return [vars[0]]
        
    @getsetDeprecated
    def _getData(self):
        return self._data

    @property
    def _data(self):
        from fipy.tools.numerix import array, reshape
        return reshape(array(self.vars[0]), self.vars[0].mesh.shape[::-1])[::-1]

    def _plot(self):
        self.norm.vmin = self._getLimit(('datamin', 'zmin'))
        self.norm.vmax = self._getLimit(('datamax', 'zmax'))
        
        self.image.set_data(self.norm(self._data))
        
        if self.colorbar is not None:
            self.colorbar.plot()

def _test():
    from fipy.viewers.viewer import _test2D
    _test2D(Matplotlib2DGridViewer)

if __name__ == "__main__": 
    import fipy.tests.doctestPlus
    fipy.tests.doctestPlus.execButNoTest()



"""
:term:`FiPy` is an object oriented, partial differential equation (PDE) solver,
written in :term:`Python`, based on a standard finite volume (FV) approach. The
framework has been developed in the Materials Science and Engineering Division
(MSED_) and Center for Theoretical and Computational Materials Science (CTCMS_),
in the Material Measurement Laboratory (MML_) at the National Institute of
Standards and Technology (NIST_).

The solution of coupled sets of PDEs is ubiquitous to the numerical
simulation of science problems.  Numerous PDE solvers exist, using a
variety of languages and numerical approaches. Many are proprietary,
expensive and difficult to customize.  As a result, scientists spend
considerable resources repeatedly developing limited tools for
specific problems.  Our approach, combining the FV method and Python_,
provides a tool that is extensible, powerful and freely available. A
significant advantage to Python_ is the existing suite of tools for
array calculations, sparse matrices and data rendering. 

The :term:`FiPy` framework includes terms for transient diffusion, convection and
standard sources, enabling the solution of arbitrary combinations of
coupled elliptic, hyperbolic and parabolic PDEs.  Currently implemented
models include phase field |citePhaseField| treatments of polycrystalline,
dendritic, and electrochemical phase transformations as well as a level set
treatment of the electrodeposition process |citeCEAC|.

.. _MML:                  http://www.nist.gov/mml/
.. _CTCMS:                http://www.ctcms.nist.gov/
.. _MSED:                 http://www.nist.gov/mml/msed/
.. _NIST:                 http://www.nist.gov/
"""
__docformat__ = 'restructuredtext'

def _getVersion():
    from pkg_resources import get_distribution, DistributionNotFound

    try:
        version = get_distribution(__name__).version
    except DistributionNotFound:
        version = "unknown, try running `python setup.py egg_info`"
        
    return version
    
__version__ = _getVersion()

from fipy.boundaryConditions import *
from fipy.meshes import *
from fipy.solvers import *
from fipy.steppers import *
from fipy.terms import *
from fipy.tools import *
from fipy.variables import *
from fipy.viewers import *

__all__ = []
__all__.extend(boundaryConditions.__all__)
__all__.extend(meshes.__all__)
__all__.extend(solvers.__all__)
__all__.extend(steppers.__all__)
__all__.extend(terms.__all__)
__all__.extend(tools.__all__)
__all__.extend(variables.__all__)
__all__.extend(viewers.__all__)

# fipy needs to export raw_input whether or not parallel

import sys
if sys.version_info >= (3, 0):
    input = input
    input_original = input

    if parallelComm.Nproc > 1:
        def mpi_input(prompt=""):
            parallelComm.Barrier()
            sys.stdout.flush() 
            if parallelComm.procID == 0:
                sys.stdout.write(prompt)
                sys.stdout.flush()
                return sys.stdin.readline()
            else:
                return ""
        input = mpi_input
        
    __all__.extend(['input', 'input_original'])
else:
    raw_input = raw_input
    raw_input_original = raw_input

    if parallelComm.Nproc > 1:
        def mpi_raw_input(prompt=""):
            parallelComm.Barrier()
            sys.stdout.flush()
            if parallelComm.procID == 0:
                sys.stdout.write(prompt)
                sys.stdout.flush()
                return sys.stdin.readline()
            else:
                return ""
        raw_input = mpi_raw_input
        
    __all__.extend(['raw_input', 'raw_input_original'])

_saved_stdout = sys.stdout

def _serial_doctest_raw_input(prompt):
    """Replacement for `raw_input()` that works in doctests
    """
    _saved_stdout.write("\n")
    _saved_stdout.write(prompt)
    _saved_stdout.flush()
    return sys.stdin.readline()

def doctest_raw_input(prompt):
    """Replacement for `raw_input()` that works in doctests
    
    This routine attempts to be savvy about running in parallel.
    """
    try:
        from fipy.tools import parallelComm
        parallelComm.Barrier()
        _saved_stdout.flush()
        if parallelComm.procID == 0:
            txt = _serial_doctest_raw_input(prompt)
        else:
            txt = ""
        parallelComm.Barrier()
    except ImportError:
        txt = _serial_doctest_raw_input(prompt)
#     return txt

def test(*args):
    r"""
    Test `Fipy`. Equivalent to::
    
    $ python setup.py test --modules

    Use

    >>> import fipy
    >>> fipy.test('--help')

    for a full list of options. Options can be passed in the same way
    as they are appended at the command line. For example, to test
    `FiPy` with `Trilinos` and inlining switched on, use

    >>> fipy.test('--trilinos', '--inline')

    At the command line this would be::

    $ python setup.py test --modules --trilinos --inline

    """

    from setuptools import setup
    from fipy.tests.testClass import _TestClass
    from setuptools.command.test import test as _test
    import tempfile

    tmpDir = tempfile.mkdtemp()

    try:
        setup(name='FiPy',
              script_args = ['egg_info', '--egg-base=' + tmpDir,
                             'test', '--modules'] + list(args),
              cmdclass={'test': _TestClass(_test)})
    except SystemExit as exitErr:
        import shutil
        shutil.rmtree(tmpDir)
        raise exitErr



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "uniformGrid1D.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #  Author: James O'Beirne <james.obeirne@gmail.com>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  
 # ###################################################################
 ##

"""
1D Mesh
"""
__docformat__ = 'restructuredtext'

from fipy.tools import numerix
from fipy.tools.numerix import MA
from fipy.tools import parallelComm

from fipy.meshes.uniformGrid import UniformGrid
from fipy.meshes.builders import _UniformGrid1DBuilder
from fipy.meshes.builders import _Grid1DBuilder
from fipy.meshes.representations.gridRepresentation import _Grid1DRepresentation
from fipy.meshes.topologies.gridTopology import _Grid1DTopology

__all__ = ["UniformGrid1D"]

class UniformGrid1D(UniformGrid):
    """
    Creates a 1D grid mesh.
    
        >>> mesh = UniformGrid1D(nx = 3)
        >>> print mesh.cellCenters
        [[ 0.5  1.5  2.5]]
         
    """
    def __init__(self, dx=1., nx=1, origin=(0,), overlap=2,
                 communicator=parallelComm, 
                 _RepresentationClass=_Grid1DRepresentation,
                 _TopologyClass=_Grid1DTopology):
                           
        super(UniformGrid1D, self).__init__(communicator=communicator,
                                            _RepresentationClass=_RepresentationClass,
                                            _TopologyClass=_TopologyClass)
                                            
        builder = _UniformGrid1DBuilder()

        origin = numerix.array(origin)
        
        self.args = {
            'dx': dx, 
            'nx': nx, 
            'origin': origin, 
            'overlap': overlap
        }
        
        self._scale = {
            'length': 1.,
            'area': 1.,
            'volume': 1.
        }
         
        builder.buildGridData([dx], [nx], overlap, communicator, origin)

        ([self.dx],
         [self.nx],
         self.dim,
         scale,
         self.globalNumberOfCells,
         self.globalNumberOfFaces,
         self.overlap,
         self.offset,
         self.numberOfVertices,
         self.numberOfFaces,
         self.numberOfCells,
         self.shape,
         self.physicalShape,
         self._meshSpacing,
         self.occupiedNodes,
         self.origin) = builder.gridData

        self._setTopology()
 
    """
    Topology set and calc
    """

    def _setTopology(self):
        self._exteriorFaces = self.facesLeft | self.facesRight
                                  
    @property
    def _interiorFaces(self):
        from fipy.variables.faceVariable import FaceVariable
        interiorFaces = FaceVariable(mesh=self, value=False)
        interiorFaces[numerix.arange(self.numberOfFaces-2) + 1] = True
        return interiorFaces
            
    @property
    def _cellToFaceOrientations(self):
        orientations = numerix.ones((2, self.numberOfCells), 'l')
        if self.numberOfCells > 0:
            orientations[0] *= -1
            orientations[0,0] = 1
        return orientations

    @property
    def _adjacentCellIDs(self):
        c1 = numerix.arange(self.numberOfFaces)
        ids = numerix.array((c1 - 1, c1))
        if self.numberOfFaces > 0:
            ids[0,0] = ids[1,0]
            ids[1,-1] = ids[0,-1]
        return ids[0], ids[1]

    @property
    def _cellToCellIDs(self):
        c1 = numerix.arange(self.numberOfCells)
        ids = MA.array((c1 - 1, c1 + 1))
        if self.numberOfCells > 0:
            ids[0,0] = MA.masked
            ids[1,-1] = MA.masked
        return ids
        
    @property
    def _cellToCellIDsFilled(self):
        ids = self._cellToCellIDs.filled()
        if self.numberOfCells > 0:
            ids[0,0] = 0
            ids[1,-1] = self.numberOfCells - 1
        return ids          

    def _getExteriorFaces(self):
        return self._exteriorFaces

    def _setExteriorFaces(self, e):
        self._exteriorFaces = e
    
    exteriorFaces = property(_getExteriorFaces, _setExteriorFaces)

    """
    Geometry set and calc
    """

    @property
    def _faceAreas(self):
        return numerix.ones(self.numberOfFaces,'d')

    @property
    def _faceCenters(self):
        return numerix.arange(self.numberOfFaces)[numerix.NewAxis, ...] * self.dx + self.origin

    @property
    def faceNormals(self):
        faceNormals = numerix.ones((1, self.numberOfFaces), 'd')
        # The left-most face has neighboring cells None and the left-most cell.
        # We must reverse the normal to make fluxes work correctly.
        if self.numberOfFaces > 0:
            faceNormals[...,0] *= -1
        return faceNormals

    @property
    def _orientedFaceNormals(self):
        return self.faceNormals

    @property
    def _cellVolumes(self):
        return numerix.ones(self.numberOfCells, 'd') * self.dx

    @property
    def _cellCenters(self):
        ccs = ((numerix.arange(self.numberOfCells)[numerix.NewAxis, ...] + 0.5) \
               * self.dx + self.origin) * self.scale['length']
        return ccs

    @property
    def _cellDistances(self):
        distances = numerix.ones(self.numberOfFaces, 'd')
        distances *= self.dx
        if len(distances) > 0:
            distances[0] = self.dx / 2.
            distances[-1] = self.dx / 2.
        return distances

    @property
    def _faceTangents1(self):
        return numerix.zeros(self.numberOfFaces, 'd')[numerix.NewAxis, ...]

    @property
    def _faceTangents2(self):
        return numerix.zeros(self.numberOfFaces, 'd')[numerix.NewAxis, ...]
    
    @property
    def _cellToCellDistances(self):
        distances = MA.zeros((2, self.numberOfCells), 'd')
        distances[:] = self.dx
        if self.numberOfCells > 0:
            distances[0,0] = self.dx / 2.
            distances[1,-1] = self.dx / 2.
        return distances

    @property
    def _cellNormals(self):
        normals = numerix.ones((1, 2, self.numberOfCells), 'd')
        if self.numberOfCells > 0:
            normals[:,0] = -1
        return normals
        
    @property
    def _cellAreas(self):
        return numerix.ones((2, self.numberOfCells), 'd')

    @property
    def _cellAreaProjections(self):
        return MA.array(self._cellNormals)

    @property
    def _cellVolumes(self):
        return numerix.ones(self.numberOfCells, 'd') * self.dx
                                                          
    """
    Scaled geometry set and calc
    """

    @property
    def _faceToCellDistanceRatio(self):
        distances = numerix.ones(self.numberOfFaces, 'd')
        distances *= 0.5
        if len(distances) > 0:
            distances[0] = 1
            distances[-1] = 1
        return distances

    @property
    def _areaProjections(self):
        return self.faceNormals
        
    @property
    def _orientedAreaProjections(self):
        return self._areaProjections
        
    @property
    def _getFaceAspectRatios(self):
        return 1. / self._cellDistances


    def _translate(self, vector):
        return UniformGrid1D(dx=self.dx, 
                             nx=self.args['nx'], 
                             origin=self.args['origin'] + numerix.array(vector),
                             overlap=self.args['overlap'])

    def __mul__(self, factor):
        return UniformGrid1D(dx=self.dx * factor,
                             nx=self.args['nx'],
                             origin=self.args['origin'] * factor,
                             overlap=self.args['overlap'])

    @property
    def _concatenableMesh(self):
        from fipy.meshes.mesh1D import Mesh1D
        return Mesh1D(vertexCoords = self.vertexCoords, 
                      faceVertexIDs = _Grid1DBuilder.createFaces(self.numberOfVertices), 
                      cellFaceIDs = _Grid1DBuilder.createCells(self.nx))
                      
    @property
    def _cellFaceIDs(self):
        return MA.array(_Grid1DBuilder.createCells(self.nx))

    @property
    def _maxFacesPerCell(self):
        return 2
    
    @property
    def vertexCoords(self):
        return numerix.array(self.faceCenters)

    @property
    def faceCellIDs(self):
        c1 = numerix.arange(self.numberOfFaces)
        ids = MA.array((c1 - 1, c1))
        if self.numberOfFaces > 0:
            ids[0,0] = ids[1,0]
            ids[1,0] = MA.masked
            ids[1,-1] = MA.masked
        return ids

    @property
    def _cellVertexIDs(self):
        c1 = numerix.arange(self.numberOfCells)
        return numerix.array((c1 + 1, c1))

    def _getNearestCellID(self, points):
        """
        Test cases

           >>> from fipy import *
           >>> m = Grid1D(nx=3)
           >>> print m._getNearestCellID(([0., .9, 3.],))
           [0 0 2]
           >>> print m._getNearestCellID(([1.1],))
           [1]
           >>> m0 = Grid1D(nx=2, dx=1.)
           >>> m1 = Grid1D(nx=4, dx=.5)
           >>> print m0._getNearestCellID(m1.cellCenters.globalValue)
           [0 0 1 1]
           
        """
        nx = self.globalNumberOfCells
        
        if nx == 0:
            return numerix.arange(0)
            
        x0, = self.cellCenters.globalValue[...,0]        
        xi, = points
        dx = self.dx
        
        i = numerix.array(numerix.rint(((xi - x0) / dx)), 'l')
        i[i < 0] = 0
        i[i > nx - 1] = nx - 1

        return i

    def _test(self):
        """
        These tests are not useful as documentation, but are here to ensure
        everything works as expected. The following was broken, now fixed.

            >>> from fipy import *
            >>> mesh = Grid1D(nx=3., dx=1.)
            >>> var = CellVariable(mesh=mesh)
            >>> DiffusionTerm().solve(var, solver=DummySolver())

        """

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



#!/usr/bin/env python

## -*-Pyth-*-
 # ###################################################################
 #  FiPy - Python-based finite volume PDE solver
 # 
 #  FILE: "faceGradVariable.py"
 #
 #  Author: Jonathan Guyer <guyer@nist.gov>
 #  Author: Daniel Wheeler <daniel.wheeler@nist.gov>
 #  Author: James Warren   <jwarren@nist.gov>
 #    mail: NIST
 #     www: http://www.ctcms.nist.gov/fipy/
 #  
 # ========================================================================
 # This software was developed at the National Institute of Standards
 # and Technology by employees of the Federal Government in the course
 # of their official duties.  Pursuant to title 17 Section 105 of the
 # United States Code this software is not subject to copyright
 # protection and is in the public domain.  FiPy is an experimental
 # system.  NIST assumes no responsibility whatsoever for its use by
 # other parties, and makes no guarantees, expressed or implied, about
 # its quality, reliability, or any other characteristic.  We would
 # appreciate acknowledgement if the software is used.
 # 
 # This software can be redistributed and/or modified freely
 # provided that any derivative works bear some notice that they are
 # derived from it, and any modified versions bear some notice that
 # they have been modified.
 # ========================================================================
 #  See the file "license.terms" for information on usage and  redistribution
 #  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #  
 # ###################################################################
 ##

__docformat__ = 'restructuredtext'

__all__ = []

from fipy.variables.faceVariable import FaceVariable
from fipy.tools import numerix
from fipy.tools import inline

class _FaceGradVariable(FaceVariable):
    """
    Test case for a vector cell variable

    >>> from fipy import *
    >>> m = Grid2D(nx=3, ny=3)
    >>> x, y = m.cellCenters
    >>> v = CellVariable(mesh=m, elementshape=(3,))
    >>> v[0] = x
    >>> v[1] = y
    >>> v[2] = x**2
    >>> v0 = CellVariable(mesh=m, value=x)
    >>> v1 = CellVariable(mesh=m, value=y)
    >>> v2 = CellVariable(mesh=m, value=x**2)
    >>> v.faceGrad.globalValue.shape
    (2, 3, 24)
    >>> print v0.faceGrad
    [[ 0.5  1.   0.5  0.5  1.   0.5  0.5  1.   0.5  0.5  1.   0.5  0.   1.   1.
       0.   0.   1.   1.   0.   0.   1.   1.   0. ]
     [ 0.   0.   0.   0.   0.   0.   0.   0.   0.   0.   0.   0.   0.   0.   0.
       0.   0.   0.   0.   0.   0.   0.   0.   0. ]]
    >>> print (v0.faceGrad.globalValue == v.faceGrad.globalValue[:,0]).all()
    True
    >>> print (v1.faceGrad.globalValue  == v.faceGrad.globalValue[:,1]).all()
    True
    >>> print (v2.faceGrad.globalValue  == v.faceGrad.globalValue[:,2]).all()
    True
     
    """
    def __init__(self, var):
        FaceVariable.__init__(self, mesh=var.mesh, elementshape=(var.mesh.dim,) + var.shape[:-1])
        self.var = self._requires(var)

    def _calcValue(self):
        if inline.doInline and self.var.rank == 0:
            return self._calcValueInline()
        else:
            return self._calcValueNoInline()

    def _calcValueInline(self):

        id1, id2 = self.mesh._adjacentCellIDs

        tangents1 = self.mesh._faceTangents1
        tangents2 = self.mesh._faceTangents2

        val = self._array.copy()

        faceNormals = self.mesh._orientedFaceNormals
        if numerix.MA.isMaskedArray(faceNormals):
            faceNormals = faceNormals.filled()

        inline._runIterateElementInline("""
            int j;
            double t1grad1, t1grad2, t2grad1, t2grad2, N, N2;
            int ID1 = ITEM(id1, i, NULL);
            int ID2 = ITEM(id2, i, NULL);

            if ITEM(exteriorFaces, i, NULL) {
                 N2 = ITEM(facevar, i, NULL);
            } else {
                 N2 = ITEM(var, ID2, NULL);
            }

            N = (N2 - ITEM(var, ID1, NULL)) / ITEM(dAP, i, NULL);

            t1grad1 = t1grad2 = t2grad1 = t2grad2 = 0.;

            t1grad1 += ITEM(tangents1, i, vec) * ITEM(cellGrad, ID1, vec);
            t1grad2 += ITEM(tangents1, i, vec) * ITEM(cellGrad, ID2, vec);
            t2grad1 += ITEM(tangents2, i, vec) * ITEM(cellGrad, ID1, vec);
            t2grad2 += ITEM(tangents2, i, vec) * ITEM(cellGrad, ID2, vec);

            ITEM(val, i, vec) =  ITEM(normals, i, vec) * N;
            ITEM(val, i, vec) += ITEM(tangents1, i, vec) * (t1grad1 + t1grad2) / 2.;
            ITEM(val, i, vec) += ITEM(tangents2, i, vec) * (t2grad1 + t2grad2) / 2.;
        """,tangents1 = tangents1,
            tangents2 = tangents2,
            cellGrad = self.var.grad.numericValue,
            normals = faceNormals,
            id1 = id1,
            id2 = id2,
            dAP = numerix.array(self.mesh._cellDistances),
            var = self.var.numericValue,
            facevar = self.var.faceValue.numericValue,
            exteriorFaces = self.mesh.exteriorFaces.numericValue,
            val = val,
            ni = tangents1.shape[1],
            shape=numerix.array(numerix.shape(tangents1)))

        return self._makeValue(value = val)

    def _calcValueNoInline(self):
        dAP = self.mesh._cellDistances
        id1, id2 = self.mesh._adjacentCellIDs

        N2 = numerix.take(self.var.value,id2, axis=-1)

        faceMask = numerix.array(self.mesh.exteriorFaces)

        ## The following conditional is required because empty
        ## indexing is not altogether functional.  This
        ## numpy.empty((0,))[[]] and this numpy.empty((0,))[...,[]]
        ## both work, but this numpy.empty((3, 0))[...,[]] is
        ## broken.

        if self.var.faceValue.shape[-1] != 0:
            s = (Ellipsis, faceMask)
        else:
            s = (faceMask,)

        N2[s] = self.var.faceValue[s]

        N = (N2 - numerix.take(self.var,id1, axis=-1)) / dAP

        normals = self.mesh._orientedFaceNormals

        tangents1 = self.mesh._faceTangents1
        tangents2 = self.mesh._faceTangents2
        cellGrad = self.var.grad.numericValue

        grad1 = numerix.take(cellGrad, id1, axis=-1)
        grad2 = numerix.take(cellGrad, id2, axis=-1)

        s = (slice(0,None,None),) + (numerix.newaxis,) * (len(grad1.shape) - 2) + (slice(0,None,None),)
        t1grad1 = numerix.sum(tangents1[s] * grad1, 0)
        t1grad2 = numerix.sum(tangents1[s] * grad2, 0)
        t2grad1 = numerix.sum(tangents2[s] * grad1, 0)
        t2grad2 = numerix.sum(tangents2[s] * grad2, 0)

        T1 = (t1grad1 + t1grad2) / 2.
        T2 = (t2grad1 + t2grad2) / 2.

        return normals[s] * N[numerix.newaxis] + tangents1[s] * T1[numerix.newaxis] + tangents2[s] * T2[numerix.newaxis]

def _test(): 
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()
    
if __name__ == "__main__": 
    _test() 

